#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


from __future__ import print_function

import itertools


class C(object):
    def __init__(self, a, b, c, d, e, f):
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        self.e = e
        self.f = f


def calledRepeatedly():
    # Avoid module variable access speed to play a role
    local_C = C

    # This is supposed to make a call to a compiled method, which is
    # being optimized separately.
    # construct_begin
    x1 = local_C("some", "random", "values", "to", "check", "call")
    x2 = local_C("some", "other", "values", "to", "check", "call")
    x3 = local_C("some", "new", "values", "to", "check", "call")
    # construct_end


for x in itertools.repeat(None, 10000):
    calledRepeatedly()

print("OK.")

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Torch JIT plugin."""

import ast
import os

from nuitka.options.Options import getModuleParameter, isStandaloneMode
from nuitka.plugins.PluginBase import NuitkaPluginBase
from nuitka.utils.FileOperations import (
    copyFileWithPermissions,
    isRelativePath,
    makeContainingPath,
)
from nuitka.utils.ModuleNames import ModuleName


def _isTorchJitScriptCall(node, torch_names, torch_jit_names, script_names):
    called = node.func

    if isinstance(called, ast.Name):
        return called.id in script_names

    if not isinstance(called, ast.Attribute) or called.attr != "script":
        return False

    called_value = called.value

    if isinstance(called_value, ast.Name):
        return called_value.id in torch_jit_names

    if not isinstance(called_value, ast.Attribute) or called_value.attr != "jit":
        return False

    called_value = called_value.value

    return isinstance(called_value, ast.Name) and called_value.id in torch_names


def _considerTorchJitImport(alias, torch_names, torch_jit_names):
    if alias.name == "torch":
        torch_names.add(alias.asname or "torch")
    elif alias.name == "torch.jit":
        if alias.asname is None:
            torch_names.add("torch")
        else:
            torch_jit_names.add(alias.asname)


def _considerTorchJitImportFrom(node, torch_jit_names, script_names):
    if node.module == "torch":
        for alias in node.names:
            if alias.name == "jit":
                torch_jit_names.add(alias.asname or "jit")
    elif node.module == "torch.jit":
        for alias in node.names:
            if alias.name == "script":
                script_names.add(alias.asname or "script")
            elif alias.name == "*":
                script_names.add("script")


def _collectTorchJitImportNames(module_tree):
    torch_names = set()
    torch_jit_names = set()
    script_names = set()

    for node in ast.walk(module_tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                _considerTorchJitImport(
                    alias=alias,
                    torch_names=torch_names,
                    torch_jit_names=torch_jit_names,
                )
        elif isinstance(node, ast.ImportFrom) and node.level == 0:
            _considerTorchJitImportFrom(
                node=node,
                torch_jit_names=torch_jit_names,
                script_names=script_names,
            )

    return torch_names, torch_jit_names, script_names


def _moduleUsesTorchJitScript(module):
    source_filename = module.getCompileTimeFilename()

    if not os.path.isfile(source_filename):
        return False

    source_code = module.getSourceCode()

    try:
        module_tree = ast.parse(source_code, source_filename)
    except SyntaxError:
        return False

    torch_names, torch_jit_names, script_names = _collectTorchJitImportNames(
        module_tree=module_tree
    )

    if not torch_names and not torch_jit_names and not script_names:
        return False

    for node in ast.walk(module_tree):
        if isinstance(node, ast.Call) and _isTorchJitScriptCall(
            node=node,
            torch_names=torch_names,
            torch_jit_names=torch_jit_names,
            script_names=script_names,
        ):
            return True

    return False


class NuitkaPluginTorchJit(NuitkaPluginBase):
    """Include source files for compiled modules using Torch JIT."""

    plugin_name = "torch-jit"
    plugin_desc = "Required by 'torch.jit' source inspection in standalone mode."
    plugin_category = "package-support"

    @staticmethod
    def isAlwaysEnabled():
        return True

    @classmethod
    def isRelevant(cls):
        return isStandaloneMode()

    def __init__(self):
        self.torch_jit_source_files = ()

    @staticmethod
    def _isTorchJitEnabled():
        return getModuleParameter(ModuleName("torch"), "disable-jit") == "no"

    def onModuleCompleteSet(self, module_set):
        source_files = []
        seen_runtime_filenames = set()

        if not self._isTorchJitEnabled():
            self.torch_jit_source_files = ()
            return

        for module in module_set:
            if not module.isCompiledPythonModule():
                continue

            if not _moduleUsesTorchJitScript(module):
                continue

            runtime_filename = module.getRunTimeFilename()

            if runtime_filename in seen_runtime_filenames:
                continue

            seen_runtime_filenames.add(runtime_filename)

            if runtime_filename.startswith("<") or not isRelativePath(runtime_filename):
                self.warning(
                    "Cannot include source code for '%s' with runtime filename '%s'."
                    % (module.getFullName(), runtime_filename)
                )
                continue

            source_files.append(
                (
                    module.getFullName().asString(),
                    module.getCompileTimeFilename(),
                    runtime_filename,
                )
            )

        self.torch_jit_source_files = tuple(source_files)

        if self.torch_jit_source_files:
            self.info(
                "Including %d source file%s for Torch JIT."
                % (
                    len(self.torch_jit_source_files),
                    "" if len(self.torch_jit_source_files) == 1 else "s",
                )
            )

    def onStandaloneDistributionFinished(self, dist_dir):
        if not self.torch_jit_source_files:
            return

        for _module_name, source_path, runtime_filename in self.torch_jit_source_files:
            dest_path = os.path.normpath(os.path.join(dist_dir, runtime_filename))

            makeContainingPath(dest_path)

            copyFileWithPermissions(
                source_path=source_path,
                dest_path=dest_path,
                target_dir=dist_dir,
            )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


# This file helps to compute a version number in source trees obtained from
# git-archive tarball (such as those provided by githubs download-from-tag
# feature). Distribution tarballs (built by setup.py sdist) and build
# directories (produced by setup.py build) will contain a much shorter file
# that just contains the computed version number.

# This file is released into the public domain.
# Generated by versioneer-0.28
# https://github.com/python-versioneer/python-versioneer

# N.B. this is a heavily truncated version for testing purposes.

"""Git implementation of _version.py."""


def get_versions():
    # We return some arbitrary version
    return {
        "version": "0+123123",
        "full-revisionid": None,
        "dirty": None,
        "error": "unable to compute version",
        "date": None,
    }


#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Run code with Nuitka compiled and put that through Valgrind."""

import os
import sys

# Find nuitka package relative to us.
sys.path.insert(
    0, os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".."))
)

# isort:start

import shutil

from nuitka.tools.testing.Common import getTempDir
from nuitka.tools.testing.Valgrind import getBinarySizes, runValgrind

input_file = sys.argv[1]

nuitka_binary = os.getenv(
    "NUITKA_BINARY", os.path.join(os.path.dirname(__file__), "../bin/nuitka")
)
nuitka_binary = os.path.normpath(nuitka_binary)

basename = os.path.basename(input_file)

tempdir = getTempDir()

output_binary = os.path.join(
    tempdir, (basename[:-3] if input_file.endswith(".py") else basename) + ".bin"
)

os.environ["PYTHONHASHSEED"] = "0"

# To make that python run well despite the "-S" flag for things that need site
# to expand sys.path.
os.environ["PYTHONPATH"] = os.pathsep.join(sys.path)

os.system(
    "%s %s --python-flag=-S --no-progressbar --output-dir=%s %s %s %s %s"
    % (
        sys.executable,
        nuitka_binary,
        tempdir,
        "--unstripped",
        "--quiet",
        os.getenv("NUITKA_EXTRA_OPTIONS", ""),
        input_file,
    )
)

if not os.path.exists(output_binary):
    sys.exit("Seeming failure of Nuitka to compile, no %r." % output_binary)

log_base = basename[:-3] if input_file.endswith(".py") else basename

if "number" in sys.argv or "numbers" in sys.argv:
    log_file = log_base + ".log"
else:
    log_file = None

log_file = log_base + ".log"

sys.stdout.flush()


ticks = runValgrind(
    None,
    "callgrind",
    [output_binary],
    include_startup=False,
    save_log_filename=log_file,
)

if "number" in sys.argv or "numbers" in sys.argv:
    sizes = getBinarySizes(output_binary)

    print("SIZE=%d" % (sizes[0] + sizes[1]))
    print("TICKS=%s" % ticks)
    print("BINARY=%s" % nuitka_binary)

    max_mem = runValgrind(None, "massif", [output_binary], include_startup=True)

    print("MEM=%s" % max_mem)

    shutil.rmtree(tempdir)
else:
    os.system("kcachegrind 2>/dev/null 1>/dev/null %s &" % log_file)

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Torch Hub plugin."""

import ast
import os

from nuitka.__past__ import basestring
from nuitka.Errors import NuitkaForbiddenImportEncounter
from nuitka.importing.Importing import locateModule
from nuitka.importing.Recursion import recurseTo
from nuitka.options.Options import isStandaloneMode
from nuitka.plugins.PluginBase import NuitkaPluginBase
from nuitka.tree.SourceHandling import readSourceCodeFromFilename
from nuitka.utils.ModuleNames import ModuleName


class NuitkaPluginTorchHub(NuitkaPluginBase):
    """Scan Torch Hub repositories for hidden imports.

    Notes:
        Torch Hub downloads source code to a cache directory and later imports
        it from there at run time. We therefore scan the cached source tree for
        static imports and include only the external dependencies that are
        discovered. The downloaded repository itself is intentionally not added
        to the distribution, because Torch Hub will source load it again.
    """

    plugin_name = "torch-hub"
    plugin_desc = "Resolve and scan Torch Hub repositories for dependency inclusion."
    plugin_category = "package-support"

    # spell-checker: ignore yolov,hubconf,getargspec

    @classmethod
    def addPluginCommandLineOptions(cls, group):
        group.add_option(
            "--torch-hub-model",
            action="append",
            dest="torch_hub_models",
            default=[],
            help="""\
Specify a Torch Hub model to scan for hidden imports. Use the form
'REPO_OR_DIR=MODEL', e.g. 'ultralytics/yolov5=yolov5s'. The repository part
may also be a local directory containing a 'hubconf.py'. Repeat this option
for multiple models. [REQUIRED]""",
        )

    @classmethod
    def isRelevant(cls):
        return isStandaloneMode()

    def __init__(self, torch_hub_models):
        self.model_specs = []
        self._located_module_cache = {}
        self._hub_dependency_names = set()
        self._hub_directories_seen = set()

        for torch_hub_model in torch_hub_models:
            try:
                model_spec = self._parseHubModelSpec(torch_hub_model)
            except ValueError:
                self.sysexit(
                    "Error, malformed value '%s' for '--torch-hub-model', use 'REPO_OR_DIR=MODEL'."
                    % torch_hub_model
                )
            else:
                self.model_specs.append(model_spec)

        if not self.model_specs:
            self.sysexit(
                "Error, '--torch-hub-model' must be specified when using '--enable-plugin=torch-hub'."
            )

    @staticmethod
    def _parseHubModelSpec(value):
        repo_spec, separator, model_name = value.partition("=")

        repo_spec = repo_spec.strip()
        model_name = model_name.strip()

        if separator != "=" or not repo_spec or not model_name:
            raise ValueError(value)

        return repo_spec, model_name

    def onCompilationStartChecks(self):
        for repo_spec, model_name in self.model_specs:
            hub_dir = self._resolveHubDirectory(repo_spec=repo_spec)

            if hub_dir in self._hub_directories_seen:
                continue

            self._hub_directories_seen.add(hub_dir)

            dependency_names = self._scanHubDirectory(
                hub_dir=hub_dir,
                repo_spec=repo_spec,
            )

            self._hub_dependency_names.update(dependency_names)

            self.info(
                "Torch Hub model '%s' from '%s' added %d dependency modules."
                % (model_name, repo_spec, len(dependency_names))
            )

    def onModuleInitialSet(self):
        included_module_names = set()
        attempted_module_names = set()

        for module_name in sorted(self._hub_dependency_names):
            for candidate_module_name in self._getCandidateRootModuleNames(module_name):
                if candidate_module_name in attempted_module_names:
                    continue

                attempted_module_names.add(candidate_module_name)

                included_module = self._includeDependencyModule(
                    module_name=candidate_module_name
                )

                if included_module is not None:
                    included_module_names.add(candidate_module_name)
                    yield included_module

    @staticmethod
    def _makeInfoName(prefix, value):
        result = []

        for c in value:
            if c.isalnum():
                result.append(c)
            else:
                result.append("_")

        return "%s_%s" % (prefix, "".join(result))

    def _resolveHubDirectory(self, repo_spec):
        hub_dir = self.queryRuntimeInformationSingle(
            info_name=self._makeInfoName("hub_dir", repo_spec),
            setup_codes="""\
import inspect
import os
import torch

def _resolveTorchHubDirectory(repo_spec):
    if os.path.isdir(repo_spec):
        return os.path.abspath(repo_spec)

    cache_or_reload = getattr(torch.hub, "_get_cache_or_reload", None)

    if cache_or_reload is None:
        raise RuntimeError("torch.hub helper '_get_cache_or_reload' is unavailable")

    try:
        argument_names = inspect.getfullargspec(cache_or_reload).args
    except AttributeError:
        argument_names = inspect.getargspec(cache_or_reload).args

    kwargs = {}

    if "force_reload" in argument_names:
        kwargs["force_reload"] = False

    if "trust_repo" in argument_names:
        kwargs["trust_repo"] = True

    if "verbose" in argument_names:
        kwargs["verbose"] = False

    if "skip_validation" in argument_names:
        kwargs["skip_validation"] = True

    return cache_or_reload(repo_spec, **kwargs)
""",
            value="_resolveTorchHubDirectory(%r)" % repo_spec,
        )

        if not hub_dir:
            return self.sysexit(
                "Error, failed to resolve Torch Hub repository '%s'." % repo_spec
            )

        hub_conf_filename = os.path.join(hub_dir, "hubconf.py")

        if not os.path.isfile(hub_conf_filename):
            return self.sysexit(
                "Error, Torch Hub repository '%s' at '%s' has no 'hubconf.py'."
                % (repo_spec, hub_dir)
            )

        return hub_dir

    def _scanHubDirectory(self, hub_dir, repo_spec):
        internal_modules = self._collectHubInternalModules(hub_dir=hub_dir)

        if "hubconf" not in internal_modules:
            return self.sysexit(
                "Error, Torch Hub repository '%s' at '%s' has no importable 'hubconf' module."
                % (repo_spec, hub_dir)
            )

        external_module_names = set()
        pending = ["hubconf"]
        processed = set()

        while pending:
            module_name = pending.pop()

            if module_name in processed:
                continue

            processed.add(module_name)

            source_filename, is_package = internal_modules[module_name]

            if os.path.isdir(source_filename):
                continue

            for dependency_name in self._readModuleDependencyNames(
                module_name=module_name,
                source_filename=source_filename,
                is_package=is_package,
                internal_modules=internal_modules,
            ):
                if dependency_name in internal_modules:
                    pending.append(dependency_name)
                else:
                    external_module_names.add(dependency_name)

        return external_module_names

    @staticmethod
    def _getHubPackageName(hub_dir, root):
        rel_dir = os.path.relpath(root, hub_dir)

        if rel_dir == ".":
            return None

        return rel_dir.replace(os.path.sep, ".")

    @staticmethod
    def _hasHubPackageContent(dirnames, filenames):
        for filename in filenames:
            if filename.endswith(".py"):
                return True

        for dirname in dirnames:
            if not dirname.startswith("."):
                return True

        return False

    @staticmethod
    def _collectHubInternalModules(hub_dir):
        result = {}

        for root, dirnames, filenames in os.walk(hub_dir):
            dirnames[:] = [
                dirname
                for dirname in dirnames
                if dirname != "__pycache__" and not dirname.startswith(".")
            ]

            package_name = NuitkaPluginTorchHub._getHubPackageName(
                hub_dir=hub_dir,
                root=root,
            )

            if package_name is not None and NuitkaPluginTorchHub._hasHubPackageContent(
                dirnames=dirnames,
                filenames=filenames,
            ):
                result[package_name] = root, True

            if package_name is not None and "__init__.py" in filenames:
                result[package_name] = os.path.join(root, "__init__.py"), True

            for filename in filenames:
                if filename == "__init__.py" or not filename.endswith(".py"):
                    continue

                module_basename = filename[:-3]

                if package_name is None:
                    module_name = module_basename
                else:
                    module_name = package_name + "." + module_basename

                result[module_name] = os.path.join(root, filename), False

        return result

    def _readModuleDependencyNames(
        self, module_name, source_filename, is_package, internal_modules
    ):
        module_name_obj = ModuleName(module_name)

        source_code = readSourceCodeFromFilename(
            module_name=module_name_obj,
            source_filename=source_filename,
            pre_load=True,
        )

        try:
            module_tree = ast.parse(source_code, source_filename)
        except SyntaxError:
            self.warning(
                "Torch Hub dependency module '%s' could not be parsed due to a SyntaxError."
                % module_name
            )
            return set()

        result = set()

        for dependency_name in self._extractHubDeclaredDependencies(module_tree):
            result.add(dependency_name)

        for node in ast.walk(module_tree):
            if isinstance(node, ast.Import):
                for name in node.names:
                    result.add(name.name)
            elif isinstance(node, ast.ImportFrom):
                imported_module_name = self._resolveImportFromModuleName(
                    module_name=module_name,
                    is_package=is_package,
                    node=node,
                )

                if imported_module_name is None:
                    continue

                result.add(imported_module_name)

                for imported_name in node.names:
                    if imported_name.name == "*":
                        continue

                    child_module_name = imported_module_name + "." + imported_name.name

                    if self._isImportableModuleName(
                        module_name=child_module_name,
                        internal_modules=internal_modules,
                    ):
                        result.add(child_module_name)

        return result

    @staticmethod
    def _extractHubDeclaredDependencies(module_tree):
        result = set()

        for node in module_tree.body:
            if not isinstance(node, ast.Assign):
                continue

            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == "dependencies":
                    for value in NuitkaPluginTorchHub._extractStringSequence(
                        node.value
                    ):
                        result.add(value)

        return result

    @staticmethod
    def _extractStringSequence(node):
        if not isinstance(node, (ast.List, ast.Tuple)):
            return ()

        result = []

        for element in node.elts:  # spell-checker: ignore elts
            element_value = NuitkaPluginTorchHub._extractStringConstant(element)

            if element_value is None:
                return ()

            result.append(element_value)

        return tuple(result)

    @staticmethod
    def _extractStringConstant(node):
        constant_type = getattr(ast, "Constant", None)

        if isinstance(node, ast.Str):
            return node.s

        if constant_type is not None and isinstance(node, constant_type):
            if isinstance(node.value, basestring):
                return node.value

        return None

    @staticmethod
    def _getCurrentPackageName(module_name, is_package):
        if is_package:
            return module_name

        if "." not in module_name:
            return None

        return module_name.rsplit(".", 1)[0]

    @staticmethod
    def _resolveImportFromModuleName(module_name, is_package, node):
        if node.level == 0:
            return node.module

        current_package_name = NuitkaPluginTorchHub._getCurrentPackageName(
            module_name=module_name,
            is_package=is_package,
        )

        if current_package_name is None:
            return node.module

        package_parts = current_package_name.split(".")

        if node.level > len(package_parts) + 1:
            return None

        prefix_parts = package_parts[: len(package_parts) - (node.level - 1)]

        if node.module:
            prefix_parts.append(node.module)

        if not prefix_parts:
            return None

        return ".".join(prefix_parts)

    def _isImportableModuleName(self, module_name, internal_modules):
        if module_name in internal_modules:
            return True

        _module_name, module_filename, _module_kind, finding = self._locateModule(
            module_name
        )

        return module_filename is not None and finding != "not-found"

    def _locateModule(self, module_name):
        if module_name not in self._located_module_cache:
            self._located_module_cache[module_name] = locateModule(
                module_name=ModuleName(module_name),
                parent_package=None,
                level=0,
            )

        return self._located_module_cache[module_name]

    @staticmethod
    def _getCandidateRootModuleNames(module_name):
        result = []
        parts = module_name.split(".")

        for count in range(1, len(parts) + 1):
            result.append(".".join(parts[:count]))

        return result

    def _includeDependencyModule(self, module_name):
        (
            resolved_module_name,
            module_filename,
            module_kind,
            finding,
        ) = self._locateModule(module_name)

        if (
            finding == "not-found"
            or module_filename is None
            and module_kind != "built-in"
        ):
            self.warning(
                "Torch Hub dependency module '%s' was not found during inclusion."
                % module_name
            )
            return None

        if module_kind == "built-in":
            return None

        try:
            return recurseTo(
                module_name=resolved_module_name,
                module_filename=module_filename,
                module_kind=module_kind,
                source_ref=None,
                reason="torch hub dependency",
                using_module_name=None,
            )
        except NuitkaForbiddenImportEncounter as e:
            return self.sysexit(
                """\
Error, forbidden import of '%s' (intending to avoid '%s') encountered while including Torch Hub dependency '%s'."""
                % (e.args[0], e.args[1], module_name)
            )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""This test runner compiles all Python files as a module.

This is a test to achieve some coverage, it will only find assertions of
within Nuitka or warnings from the C compiler. Code will not be run
normally.

"""

import os
import sys

# Find nuitka package relative to us.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
    ),
)

# isort:start

from nuitka import Options
from nuitka.__past__ import iter_modules
from nuitka.importing.Importing import (
    addMainScriptDirectory,
    decideModuleSourceRef,
    locateModule,
)
from nuitka.tools.testing.Common import my_print, setup, test_logger
from nuitka.Tracing import plugins_logger
from nuitka.tree.SourceHandling import (
    getSourceCodeDiff,
    readSourceCodeFromFilenameWithInformation,
)
from nuitka.utils.ModuleNames import ModuleName

python_version = setup(suite="python_modules", needs_io_encoding=True)


addMainScriptDirectory("/doesnotexist")
Options.is_full_compat = False


def scanModule(name_space, module_iterator):
    # plenty details here, pylint: disable=too-many-branches,too-many-locals

    from nuitka.tree.TreeHelpers import parseSourceCodeToAst

    for module_desc in module_iterator:
        if name_space is None:
            module_name = ModuleName(module_desc.name)
        else:
            module_name = name_space.getChildNamed(module_desc.name)

        try:
            _module_name, module_filename, finding = locateModule(
                module_name=module_name, parent_package=None, level=0
            )
        except AssertionError:
            # TODO: Currently bytecode only modules are triggering an assertion.
            continue

        assert _module_name == module_name, module_desc

        # Only source code that is found.
        if module_filename is None:
            continue

        (
            _main_added,
            _is_package,
            _is_namespace,
            _source_ref,
            source_filename,
        ) = decideModuleSourceRef(
            filename=module_filename,
            module_name=module_name,
            is_main=False,
            is_fake=False,
            logger=test_logger,
        )

        try:
            (
                source_code,
                original_source_code,
                contributing_plugins,
            ) = readSourceCodeFromFilenameWithInformation(
                module_name=module_name, source_filename=source_filename
            )
        except SyntaxError:
            continue

        try:
            parseSourceCodeToAst(
                source_code=source_code,
                module_name=module_name,
                filename=source_filename,
                line_offset=0,
            )
        except (SyntaxError, IndentationError) as e:
            try:
                parseSourceCodeToAst(
                    source_code=original_source_code,
                    module_name=module_name,
                    filename=source_filename,
                    line_offset=0,
                )
            except (SyntaxError, IndentationError):
                # Also an exception without the plugins, that is OK
                pass
            else:
                source_diff = getSourceCodeDiff(original_source_code, source_code)

                for line in source_diff:
                    plugins_logger.warning(line)

                if len(contributing_plugins) == 1:
                    contributing_plugins[0].sysexit(
                        "Making changes to '%s' that cause SyntaxError '%s'"
                        % (module_name, e)
                    )
                else:
                    test_logger.sysexit(
                        "One of the plugins '%s' is making changes to '%s' that cause SyntaxError '%s'"
                        % (",".join(contributing_plugins), module_name, e)
                    )

        my_print(module_name, ":", finding, "OK")

        if module_desc.ispkg:
            scanModule(module_name, iter_modules([module_filename]))


def main():
    scanModule(None, iter_modules())


if __name__ == "__main__":
    main()

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Many tools work on Nuitka sources and need to find the files."""

import os

from nuitka.utils.Shebang import getShebangFromFile

_default_ignore_list = ("inline_copy", "tblib", "__pycache__")


def _isIgnoredDirectoryName(dirname):
    """Decide if a directory should be ignored while scanning sources."""

    return (
        dirname in _default_ignore_list
        or dirname.endswith((".build", ".onefile-build", ".dist", ".app", ".egg-info"))
        or dirname.startswith(("CPython", "venv_", ".venv_", "env_", ".env_"))
    )


def _addFromDirectory(path, suffixes, ignore_list):
    """Recursively find files in a directory with specific suffixes.

    Args:
        path: str - directory path to start from
        suffixes: tuple - file suffixes to match
        ignore_list: tuple - list of filenames to ignore
    """
    for dirpath, dirnames, filenames in os.walk(path):
        dirnames.sort()

        # Remove things we never care about.
        dirnames[:] = [
            dirname for dirname in dirnames if not _isIgnoredDirectoryName(dirname)
        ]

        filenames.sort()

        for filename in filenames:
            if filename in ignore_list:
                continue

            fullpath = os.path.join(dirpath, filename)

            # Ignore links
            if os.path.islink(fullpath):
                continue

            # Skip temporary files from flymake mode of Emacs, spell-checker: ignore flymake
            if filename.endswith("_flymake.py"):
                continue
            # Skip temporary files from unsaved files of Emacs.
            if filename.startswith(".#"):
                continue
            # Skip bytecode files
            if filename.endswith((".pyc", ".pyo")):
                continue
            # Skip executables
            if filename.endswith((".exe", ".bin")):
                continue

            # Python files only might include files with a shebang that points
            # to Python.
            if ".py" in suffixes and not filename.endswith(suffixes):
                shebang = getShebangFromFile(fullpath)

                if shebang is None or "python" not in shebang:
                    continue

            yield fullpath


def scanTargets(positional_args, suffixes, ignore_list=()):
    """Scan list of paths for files with specific suffixes.

    Args:
        positional_args: list - paths to scan (files or directories)
        suffixes: tuple - file suffixes to match
        ignore_list: tuple - list of filenames to ignore
    """
    for positional_arg in positional_args:
        positional_arg = os.path.normpath(positional_arg)

        if os.path.isdir(positional_arg):
            if _isIgnoredDirectoryName(os.path.basename(positional_arg)):
                continue

            for value in _addFromDirectory(positional_arg, suffixes, ignore_list):
                yield value
        else:
            yield positional_arg


def isPythonFile(filename, effective_filename=None):
    """Check if a file is a Python file.

    Checks extension and shebang for files without extension.

    Args:
        filename: str - path to the file
        effective_filename: str - effective filename to use for extension check
    """
    if effective_filename is None:
        effective_filename = filename

    if effective_filename.lower().endswith((".py", ".pyw", ".scons")):
        return True
    elif os.path.isdir(filename):
        return False
    else:
        shebang = getShebangFromFile(filename)

        if shebang is not None:
            shebang = shebang[2:].lstrip()
            if shebang.startswith("/usr/bin/env"):
                shebang = shebang[12:].lstrip()

            if shebang.startswith("python"):
                return True

    return False


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Internal tool, attach the standalone distribution in compressed form."""

import os
import sys

if __name__ == "__main__":
    sys.path.insert(0, os.environ["NUITKA_PACKAGE_HOME"])

    import nuitka  # just to have it loaded from there, pylint: disable=unused-import

    del sys.path[0]

    sys.path = [
        path_element
        for path_element in sys.path
        if os.path.dirname(os.path.abspath(__file__)) != path_element
    ]

    from nuitka.tools.onefile_compressor.OnefileCompressor import main

    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


import itertools

module_value1 = 5
module_value2 = 3


def calledRepeatedly():
    # Make sure we have a local variable x anyway
    x = module_value2

    local_value = module_value1

    # construct_begin
    for x in range(local_value, local_value + 3):
        pass
    # construct_end

    return x


for x in itertools.repeat(None, 50000):
    calledRepeatedly()

print("OK.")

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


# Show relative speeds of local, nonlocal, global, and built-in access.

trials = [1] * 500


def read_local(trials=trials):
    v_local = 1
    for t in trials:
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local
        v_local


def make_nonlocal_reader():
    v_nonlocal = 1

    def inner(trials=trials):
        for t in trials:
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal
            v_nonlocal

    inner.__name__ = "read_nonlocal"
    return inner


read_nonlocal = make_nonlocal_reader()

v_global = 1


def read_global(trials=trials):
    for t in trials:
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global
        v_global


def read_builtin(trials=trials):
    for t in trials:
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct
        oct


class A(object):
    def m(self):
        pass


def read_classvar(trials=trials, A=A):
    A.x = 1
    for t in trials:
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x
        A.x


def read_instancevar(trials=trials, a=A()):
    a.x = 1
    for t in trials:
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x
        a.x


def read_unboundmethod(trials=trials, A=A):
    # real unbound methods are only in Py2.x
    for t in trials:
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m
        A.m


def read_boundmethod(trials=trials, a=A()):
    for t in trials:
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m
        a.m


def write_local(trials=trials):
    v_local = 1
    for t in trials:
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1
        v_local = 1


def make_nonlocal_writer():
    v_nonlocal = 1

    def inner(trials=trials):
        # nonlocal v_nonlocal            # this is invalid syntax in Py2.x
        for t in trials:
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1
            v_nonlocal = 1

    inner.__name__ = "write_nonlocal"
    return inner


write_nonlocal = make_nonlocal_writer()


def write_global(trials=trials):
    global v_global
    for t in trials:
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1
        v_global = 1


def write_classvar(trials=trials, A=A):
    for t in trials:
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1
        A.x = 1


def write_instancevar(trials=trials, a=A()):
    for t in trials:
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1
        a.x = 1


def loop_overhead(trials=trials):
    for t in trials:
        pass


if __name__ == "__main__":
    from timeit import Timer

    for f in [
        read_local,
        read_nonlocal,
        read_global,
        read_builtin,
        read_classvar,
        read_instancevar,
        read_unboundmethod,
        read_boundmethod,
        write_local,
        write_nonlocal,
        write_global,
        write_classvar,
        write_instancevar,
        loop_overhead,
    ]:
        print("{:5.3f}\t{}".format(min(Timer(f).repeat(7, 1000)), f.__name__))

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


import itertools
import sys

module_value1 = "1000"
module_value2 = "2000"

loop_count = 50000 if len(sys.argv) < 2 else int(sys.argv[1])

def calledRepeatedly(cond, value, str1):
    if cond:
        if str1 == "1000":
            value = not value
    else:
        value = not value

    return value

def makeSame(value):
    new_value = "".join(list(value))
    assert new_value is not value

module_value1_copy = makeSame(module_value1)

for x in range(10000):
    # construct_begin
    calledRepeatedly(True, x, module_value1)
    calledRepeatedly(True, x, module_value1_copy)
    # construct_alternative
    calledRepeatedly(False, x, module_value1)
    calledRepeatedly(False, x, module_value1_copy)
    # construct_end


print("OK.")

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2020, Tommy Li, mailto:<tommyli3318@gmail.com> find license text at end of file


"""Runner for PyPI Pytest comparison

This script automates the comparing of pytest results of a nuitka compiled wheel
using `python setup.py bdist_nuitka` to the pytest results of an uncompiled wheel
built using `python setup.py bdist_wheel` for the most popular PyPI packages.
Testing is done to ensure that nuitka is building the wheel correctly. If the
pytest run passes/fails in the same way, that means Nuitka built the wheel properly.
Else if the tests differ, then something is wrong.
Virtualenv is used to create a clean environment with no outside pollution.

"""

import json
import os
import sys

# Find nuitka package relative to us.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
    ),
)

# isort:start

import nuitka
from nuitka.tools.environments.Virtualenv import withVirtualenv
from nuitka.tools.testing.Common import (
    createSearchMode,
    my_print,
    reportSkip,
    setup,
    test_logger,
)
from nuitka.tools.testing.OutputComparison import compareOutput
from nuitka.utils.AppDirs import getCacheDir
from nuitka.utils.FileOperations import getFileContents


def executeCommand(command):
    test_logger.info("Executing: %s" % command, style="blue")

    return os.system(command) == 0


def gitClone(package, url, directory):
    """
    Update package with git if already existing in directory
    else git clone the package into directory
    """

    os.chdir(directory)
    if not executeCommand(
        "cd %s && git fetch -q && git reset -q --hard origin && git clean -q -dfx"
        % package
    ):
        assert executeCommand(
            "git clone %s %s --depth 1 --single-branch --no-tags" % (url, package)
        ), ("Error while git cloning package %s, aborting..." % package)


def main():
    # pylint: disable=broad-except,too-many-branches,too-many-locals,too-many-statements

    setup(suite="pypi")

    # cache_dir is where the git clones are cached
    cache_dir = getCacheDir("pypi-git-clones")
    base_dir = os.getcwd()

    if not os.path.isdir(cache_dir):
        os.mkdir(cache_dir)

    search_mode = createSearchMode()

    results = []

    # load json
    packages = json.load(getFileContents("packages.json"))

    for package_name, details in sorted(packages.items()):
        active = search_mode.consider(dirname=None, filename=package_name)

        if not active:
            continue

        if str is not bytes:
            # running on python3
            if package_name in ("futures", "future"):
                reportSkip("Does not run on Python3", ".", package_name)
                continue

        if os.name == "nt":
            if package_name in ("cryptography",):
                reportSkip("Not working on Windows", ".", package_name)
                continue

        if package_name == "pyyaml":
            reportSkip("Not yet supported, see Issue #476", ".", package_name)
            continue

        if package_name in ("pycparser", "numpy"):
            reportSkip("Not yet supported, see Issue #477", ".", package_name)
            continue

        if package_name in (
            "google-auth",  # bdist_nuitka fails AttributeError: single_version_externally_managed
            "jinja2",  # ModuleNotFoundError: No module named 'jinja2.tests'
            "pandas",  # ModuleNotFoundError: No module named 'Cython'
            "pytz",  # need to 'make build'
            "rsa",  # Now uses Poetry (no setup.py)
        ):
            continue

        package_dir = os.path.join(cache_dir, package_name)

        try:
            gitClone(package_name, details["url"], cache_dir)

            os.chdir(base_dir)
            with withVirtualenv(
                "venv_%s" % package_name, logger=test_logger, delete=False, style="blue"
            ) as venv:
                dist_dir = os.path.join(package_dir, "dist")

                # delete ignored tests if any
                if details["ignored_tests"]:
                    for test in details["ignored_tests"]:
                        venv.runCommand("rm -rf %s" % os.path.join(package_dir, test))

                # setup for pytest
                cmds = [
                    "python -m pip install pytest",
                    "cd %s" % os.path.join(os.path.dirname(nuitka.__file__), ".."),
                    "python setup.py develop",
                    "cd %s" % package_dir,
                ]

                if details["requirements_file"]:
                    cmds.append(
                        "python -m pip install -r %s" % details["requirements_file"]
                    )

                if details.get("extra_commands"):
                    cmds += details["extra_commands"]

                # build uncompiled .whl
                cmds.append("python setup.py bdist_wheel")

                venv.runCommand(commands=cmds)

                # install and print out if the active .whl is compiled or not
                venv.runCommand(
                    commands=[
                        "python -m pip install -U %s"
                        % os.path.join(dist_dir, os.listdir(dist_dir)[0]),
                        # use triple quotes for linux
                        """python -c "print(getattr(__import__('%s'),'__compiled__','__uncompiled_version__'))" """
                        % details.get("package_name", package_name),
                    ]
                )

                # get uncompiled pytest results
                uncompiled_stdout, uncompiled_stderr = venv.runCommandWithOutput(
                    commands=[
                        "cd %s" % package_dir,
                        "python -m pytest --disable-warnings",
                    ],
                    style="blue",
                )

                # clean up before building compiled .whl
                cmds = ["cd %s" % package_dir, "git clean -dfx"]

                if details.get("extra_commands"):
                    cmds += details["extra_commands"]

                # build nuitka compiled .whl
                cmds.append("python setup.py bdist_nuitka")

                venv.runCommand(commands=cmds)

                # install and print out if the active .whl is compiled or not
                venv.runCommand(
                    commands=[
                        "python -m pip install -U %s"
                        % os.path.join(dist_dir, os.listdir(dist_dir)[0]),
                        # use triple quotes for linux
                        """python -c "print(getattr(__import__('%s'),'__compiled__','__uncompiled_version__'))" """
                        % details.get("package_name", package_name),
                    ]
                )

                # get compiled pytest results, may fail some tests.
                (
                    compiled_stdout,
                    compiled_stderr,
                    _exit_code,
                ) = venv.runCommandWithOutput(
                    commands=[
                        "cd %s" % package_dir,
                        "python -m pytest --disable-warnings",
                    ],
                    style="blue",
                )

                venv.runCommand(commands=["cd %s" % package_dir, "git clean -q -dfx"])

        except Exception as e:
            test_logger.warning(
                "Package",
                package_name,
                "ran into an exception during execution, traceback: ",
            )
            test_logger.info(repr(e))
            results.append((package_name, "ERROR", "ERROR"))

            continue

        # compare outputs
        stdout_diff = compareOutput(
            "stdout",
            uncompiled_stdout,
            compiled_stdout,
            ignore_warnings=True,
            syntax_errors=True,
        )

        stderr_diff = compareOutput(
            "stderr",
            uncompiled_stderr,
            compiled_stderr,
            ignore_warnings=True,
            syntax_errors=True,
        )

        results.append((package_name, stdout_diff, stderr_diff))

        exit_code = stdout_diff or stderr_diff

        my_print(
            "\n=================================================================================",
            "\n--- %s ---" % package_name,
            "exit_stdout:",
            stdout_diff,
            "exit_stderr:",
            stderr_diff,
            (
                "\nError, outputs differed for package %s." % package_name
                if exit_code
                else "\nNo differences found for package %s." % package_name
            ),
            "\n=================================================================================\n",
            style="red" if exit_code else "green",
        )

        if exit_code != 0 and search_mode.abortOnFinding(
            dirname=None, filename=package_name
        ):
            break

    search_mode.finish()

    # give a summary of all packages
    my_print(
        "\n\n=====================================SUMMARY=====================================",
        style="yellow",
    )

    for package_name, stdout_diff, stderr_diff in results:
        my_print(
            package_name,
            "-",
            end=" ",
            style="red" if (stdout_diff or stderr_diff) else "green",
        )

        my_print(
            "stdout:", stdout_diff, end=" ", style="red" if stdout_diff else "green"
        )

        my_print(
            "stderr:", stderr_diff, end="", style="red" if stderr_diff else "green"
        )

        my_print(
            "\n---------------------------------------------------------------------------------"
        )

    my_print("TOTAL NUMBER OF PACKAGES TESTED: %s" % len(results), style="yellow")

    num_failed = 0
    num_errors = 0

    # tally the number of errors and failed
    for _, y, z in results:
        if type(y) is str:
            # this means the package ran into an exception
            num_errors += 1
        elif y or z:
            num_failed += 1

    my_print(
        "TOTAL PASSED: %s" % (len(results) - num_failed - num_errors), style="green"
    )

    my_print("TOTAL FAILED (differences): %s" % num_failed, style="red")

    my_print("TOTAL ERRORS (exceptions): %s" % num_errors, style="red")


if __name__ == "__main__":
    main()

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Tests that reads a data file via path"""

# nuitka-project: --include-package=some_package

from importlib.resources import contents, is_resource, path

# Test contents (returns iterable of resource names)
print("CONTENTS", sorted(contents("some_package")))

# Test is_resource
assert is_resource("some_package", "DATA_FILE.txt") is True
assert is_resource("some_package", "missing.txt") is False

# Test path
with path("some_package", "DATA_FILE.txt") as data_path:
    with open(data_path, encoding="utf8") as data_file:
        print("RES", data_file.read())

print("OK.")

#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Main program for PyLint checker tool."""

import os
import time

from nuitka.options.CommandLineOptionsTools import makeOptionsParser
from nuitka.PythonVersions import python_version
from nuitka.tools.Basics import addPYTHONPATH, getHomePath, setupPATH
from nuitka.tools.quality.Git import addGitArguments, getGitPaths
from nuitka.tools.quality.pylint.PyLint import executePyLint
from nuitka.tools.quality.ScanSources import isPythonFile, scanTargets
from nuitka.tools.testing.Common import hasModule, setup
from nuitka.Tracing import my_print, tools_logger
from nuitka.utils.FileOperations import (
    getFileModificationTime,
    resolveShellPatternToFilenames,
)


def _getPathParts(filename):
    filename = os.path.normpath(filename)

    if not os.path.isabs(filename):
        filename = os.path.abspath(filename)

    try:
        filename = os.path.relpath(filename, getHomePath())
    except ValueError:
        pass

    return filename.replace("\\", "/").split("/")


def isIgnoredFile(filename):
    path_parts = _getPathParts(filename)

    if path_parts[0].startswith("Mini"):
        return True
    if path_parts[0] == "examples":
        return True
    if path_parts[0] == "tests" and path_parts[-1] != "run_all.py":
        return True
    if "inline_copy" in path_parts:
        return True

    return False


def _resolveFilenames(options, positional_args):
    positional_args = getGitPaths(
        options=options,
        positional_args=positional_args,
        default_positional_args=(
            "bin",
            "nuitka",
            "setup.py",
            "tests/*/run_all.py",
        ),
    )

    if options.verbose:
        my_print("Working on: %s" % " ".join(positional_args))

    positional_args = sum(
        (
            resolveShellPatternToFilenames(positional_arg)
            for positional_arg in positional_args
        ),
        [],
    )

    ignore_list = []

    # Avoid checking the Python2 runner along with the one for Python3, it has name collisions.
    if python_version >= 0x300:
        ignore_list.append("nuitka")

    filenames = list(
        scanTargets(
            positional_args, suffixes=(".py", ".scons"), ignore_list=ignore_list
        )
    )

    # TODO: Filter this during scanTargets and during getGitPaths already
    filenames = [
        filename
        for filename in filenames
        if isPythonFile(filename)
        if not isIgnoredFile(filename)
    ]

    return filenames


def _parseArguments():
    parser = makeOptionsParser(usage=None, epilog=None)

    addGitArguments(parser)

    parser.add_option(
        "--show-todo",
        "--todo",
        action="store_true",
        dest="todo",
        default=False,
        help="""\
Show TODO items. Default is %default.""",
    )

    parser.add_option(
        "--verbose",
        action="store_true",
        dest="verbose",
        default=False,
        help="""\
Be verbose in output. Default is %default.""",
    )

    parser.add_option(
        "--one-by-one",
        action="store_true",
        dest="one_by_one",
        default=False,
        help="""\
Check files one by one. Default is %default.""",
    )

    parser.add_option(
        "--not-installed-is-no-error",
        action="store_true",
        dest="not_installed_is_no_error",
        default=False,
        help="""\
Insist on PyLint to be installed. Default is %default.""",
    )

    parser.add_option(
        "--watch",
        action="store_true",
        dest="watch",
        default=False,
        help="""\
Watch files for changes. Default is %default.""",
    )

    options, positional_args = parser.parse_args()

    return options, positional_args


def main():
    setup(go_main=False)

    # So PyLint finds nuitka package.
    addPYTHONPATH(getHomePath())
    setupPATH()

    options, positional_args = _parseArguments()

    if options.not_installed_is_no_error and not hasModule("pylint"):
        tools_logger.warning(
            "PyLint is not installed for this interpreter version: SKIPPED",
            style="yellow",
        )

        return tools_logger.sysexit(exit_code=0)

    # Scan files for changes periodically, for use in Visual Code task to
    # present errors.
    if options.watch:
        prev_filenames = set()
        prev_modification_times = {}

        while True:
            try:
                filenames = _resolveFilenames(
                    options=options, positional_args=positional_args
                )

                # Check for changes
                new_filenames = set(filenames)
                new_modification_times = {}

                for filename in filenames:
                    try:
                        new_modification_times[filename] = getFileModificationTime(
                            filename
                        )
                    except OSError:
                        new_modification_times[filename] = None

                changed = new_filenames != prev_filenames
                if not changed:
                    for filename, modification_time in new_modification_times.items():
                        if (
                            filename not in prev_modification_times
                            or prev_modification_times[filename] != modification_time
                        ):
                            changed = True
                            break

                if changed:
                    my_print(">>> Pylint Start")
                    try:
                        executePyLint(
                            filenames=filenames,
                            show_todo=options.todo,
                            verbose=options.verbose,
                            one_by_one=options.one_by_one,
                        )
                    except SystemExit:
                        pass

                    my_print(">>> Pylint End")

                    prev_filenames = new_filenames
                    prev_modification_times = new_modification_times

                time.sleep(0.5)

            except KeyboardInterrupt:
                break

        return tools_logger.sysexit(exit_code=0)

    filenames = _resolveFilenames(options=options, positional_args=positional_args)

    if not filenames:
        tools_logger.info("No matching files found that require PyLint.")
        return 0

    exit_code = executePyLint(
        filenames=filenames,
        show_todo=options.todo,
        verbose=options.verbose,
        one_by_one=options.one_by_one,
    )

    return tools_logger.sysexit(exit_code=exit_code)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Sandbox CPython's internal headers by stripping inline functions.

This ensures Nuitka modules only consume dynamically safe abstractions.
"""

import os
import re

from nuitka.options.Options import (
    isExperimental,
    isMingw64,
    shallMakeModule,
    shallUsePythonDebug,
)
from nuitka.PythonFlavors import isSelfCompiledPythonUninstalled
from nuitka.PythonVersions import (
    getTargetPythonIncludePath,
    isPythonWithGil,
    python_version,
    python_version_str,
)
from nuitka.Tracing import general
from nuitka.utils.AppDirs import getCacheDir
from nuitka.utils.Diffs import getUnifiedDiff
from nuitka.utils.FileOperations import (
    copyFile,
    getFileContents,
    getFileList,
    makePath,
    putTextFileContents,
)
from nuitka.utils.Hashing import Hash
from nuitka.utils.Json import loadJsonFromFilename
from nuitka.utils.Utils import getArchitecture, getOS, isLinux, isWin32Windows
from nuitka.Version import getNuitkaVersion


def getOffsetsJsonRequiredKeys(for_python_version_str):
    """Returns list of names we need to know the offset of."""
    # spell-checker: ignore stoptheworld,gilstate,ceval

    for_python_version_tuple = tuple(int(d) for d in for_python_version_str.split("."))
    keys = []
    if for_python_version_tuple >= (3, 12):
        keys.extend(["imports", "static_objects", "ceval"])
        if for_python_version_tuple >= (3, 13):
            keys.append("stoptheworld")
        if for_python_version_tuple >= (3, 14):
            keys.append("ref_tracer")
    else:
        keys.append("gilstate")

    return keys


def isOffsetsJsonOutdated(json_path, for_python_version_str):
    if not os.path.exists(json_path):
        return True

    data = loadJsonFromFilename(json_path)
    if data is None:
        return True

    expected_keys = set(
        "_PyRuntimeState_" + k
        for k in getOffsetsJsonRequiredKeys(for_python_version_str)
    )

    if set(data.keys()) != expected_keys:
        return True

    return False


def _replaceExpectedInAdaptedPythonHeaderFile(content, filename, old, new):
    count = content.count(old)

    assert count == 1, (
        "Expected exactly one adapted header replacement match in '%s', but got %d for %r."
        % (filename, count, old)
    )

    return content.replace(old, new)


def _substituteExpectedInAdaptedPythonHeaderFile(
    content, filename, pattern, replacement
):
    result, count = re.subn(pattern, replacement, content)

    assert count == 1, (
        "Expected exactly one adapted header regex replacement match in '%s', but got %d for %r."
        % (filename, count, pattern)
    )

    return result


def _applyExpectedAdaptedPythonHeaderReplacements(adapted, filename):
    if filename == "pycore_global_objects.h":
        adapted = _replaceExpectedInAdaptedPythonHeaderFile(
            content=adapted,
            filename=filename,
            old="(*((_PyRuntimeState*)Nuitka_Error_DoNotUseFunction())).static_objects.NAME",
            new="Nuitka_PyRuntime__static_objects->NAME",
        )

        if 0x3E0 <= python_version < 0x3F0 and shallMakeModule():
            adapted = _substituteExpectedInAdaptedPythonHeaderFile(
                content=adapted,
                filename=filename,
                pattern=(
                    r"#define _Py_INTERP_CACHED_OBJECT\(interp, NAME\) \\\n\s+"
                    r"\(interp\)->cached_objects\.NAME"
                ),
                replacement=(
                    "#define _Py_INTERP_CACHED_OBJECT(interp, NAME) \\\n"
                    "    Nuitka_PyInterpreterState_GetCachedObjects(interp)->NAME"
                ),
            )
            adapted = _substituteExpectedInAdaptedPythonHeaderFile(
                content=adapted,
                filename=filename,
                pattern=(
                    r"#define _Py_INTERP_STATIC_OBJECT\(interp, NAME\) \\\n\s+"
                    r"\(interp\)->static_objects\.NAME"
                ),
                replacement=(
                    "#define _Py_INTERP_STATIC_OBJECT(interp, NAME) \\\n"
                    "    Nuitka_PyInterpreterState_GetStaticObjects(interp)->NAME"
                ),
            )

    if filename == "pycore_object.h" and 0x3E0 <= python_version < 0x3F0:
        adapted = _substituteExpectedInAdaptedPythonHeaderFile(
            content=adapted,
            filename=filename,
            pattern=r"#\s+define _Py_DEC_REFTOTAL\(interp\) \\\n\s+interp->object_state\.reftotal--",
            replacement="#  define _Py_DEC_REFTOTAL(interp) \\\n    _Py_DecRefTotal(_PyThreadState_GET())",
        )

        if shallMakeModule() and isPythonWithGil():
            adapted = _replaceExpectedInAdaptedPythonHeaderFile(
                content=adapted,
                filename=filename,
                old="#  define _PyObject_GC_TRACK(op) \\\n        _PyObject_GC_TRACK(_PyObject_CAST(op))",
                new="#  define _PyObject_GC_TRACK(op) \\\n        Nuitka_GC_Track(_PyObject_CAST(op))",
            )
            adapted = _replaceExpectedInAdaptedPythonHeaderFile(
                content=adapted,
                filename=filename,
                old="#  define _PyObject_GC_UNTRACK(op) \\\n        _PyObject_GC_UNTRACK(_PyObject_CAST(op))",
                new="#  define _PyObject_GC_UNTRACK(op) \\\n        Nuitka_GC_UnTrack(_PyObject_CAST(op))",
            )
            adapted = _replaceExpectedInAdaptedPythonHeaderFile(
                content=adapted,
                filename=filename,
                old="""\
#  define _PyObject_GC_TRACK(op) \\
        _PyObject_GC_TRACK(__FILE__, __LINE__, _PyObject_CAST(op))""",
                new="#  define _PyObject_GC_TRACK(op) \\\n        Nuitka_GC_Track(_PyObject_CAST(op))",
            )
            adapted = _replaceExpectedInAdaptedPythonHeaderFile(
                content=adapted,
                filename=filename,
                old="""\
#  define _PyObject_GC_UNTRACK(op) \\
        _PyObject_GC_UNTRACK(__FILE__, __LINE__, _PyObject_CAST(op))""",
                new="#  define _PyObject_GC_UNTRACK(op) \\\n        Nuitka_GC_UnTrack(_PyObject_CAST(op))",
            )

    if filename == "pycore_stackref.h" and 0x3E0 <= python_version < 0x3F0:
        adapted = _substituteExpectedInAdaptedPythonHeaderFile(
            content=adapted,
            filename=filename,
            pattern=(
                r"(int tag = ref\.bits & Py_TAG_BITS;\n\s+)"
                r"PyObject \*obj = BITS_TO_PTR_MASKED\(ref\);"
            ),
            replacement=(
                r"\1NUITKA_MAY_BE_UNUSED PyObject *obj = BITS_TO_PTR_MASKED(ref);"
            ),
        )

    return adapted


def adaptPythonHeaderFile(content, filename):
    """Strip dangerous inline functions from a CPython header."""
    # A simple brace-matching state machine to rip out `static inline` blocks
    # that touch `_PyRuntime(State)`.

    lines = content.splitlines(True)
    out_lines = []

    in_inline_function = False
    brace_depth = 0
    current_block = []
    is_dangerous = False

    for line in lines:
        if not in_inline_function:
            # Detect start of static inline function
            if line.strip().startswith("static inline ") or line.strip().startswith(
                "Py_LOCAL_INLINE("
            ):
                in_inline_function = True
                brace_depth = 0
                current_block = [line]
                is_dangerous = False

                # Check for braces on the same line
                brace_depth += line.count("{")
                brace_depth -= line.count("}")
                continue

            out_lines.append(line)
        else:
            current_block.append(line)
            brace_depth += line.count("{")
            brace_depth -= line.count("}")

            # Simple heuristic detection within the block
            if any(
                t in line
                for t in (
                    "_PyRuntime",
                    "_PyLong_SMALL_INTS",
                    "_Py_tracemalloc_config",
                    "_Py_SINGLETON",
                )
            ):
                is_dangerous = True

            if brace_depth == 0 and any("{" in b for b in current_block):
                # End of function
                in_inline_function = False

                full_decl = " ".join(current_block)
                match = re.search(r"([_a-zA-Z0-9]+)\s*\(", full_decl)
                func_name = match.group(1) if match else "UNKNOWN_FUNC"

                if func_name == "_Py_freelists_GET":
                    is_dangerous = True

                if is_dangerous:
                    # Strip it entirely, replace with an error poison string
                    out_lines.append(
                        "/* Nuitka: Stripped dangerous inline function */\n"
                    )

                    out_lines.append(
                        "#define %s(...) Nuitka_Error_DoNotUseFunction()\n" % func_name
                    )
                else:
                    out_lines.extend(current_block)

    adapted = "".join(out_lines)
    # Poison any remaining direct property accesses to _PyRuntime by evaluating to an undefined function
    adapted = re.sub(
        r"(?<![a-zA-Z0-9_])_PyRuntime\.",
        "(*((_PyRuntimeState*)Nuitka_Error_DoNotUseFunction())).",
        adapted,
    )
    adapted = re.sub(
        r"&_PyRuntime(?![a-zA-Z0-9_])",
        "(&(*((_PyRuntimeState*)Nuitka_Error_DoNotUseFunction())))",
        adapted,
    )
    adapted = _applyExpectedAdaptedPythonHeaderReplacements(
        adapted=adapted, filename=filename
    )

    if "Nuitka_Error_DoNotUseFunction" in adapted:
        return "extern void * Nuitka_Error_DoNotUseFunction(void);\n" + adapted

    return adapted


def ensurePythonInternalsOffsets(cache_dir):
    """Ensure dynamic offsets are available, generating them to cache if missing."""

    if not isWin32Windows() or (python_version < 0x3D0 and not shallMakeModule()):
        return None

    keys = getOffsetsJsonRequiredKeys(python_version_str)
    if not keys:
        return None  # No offsets needed

    python_version_id = "%s-%s-%s-%s" % (
        python_version_str,
        getOS(),
        getArchitecture(),
        "gil" if isPythonWithGil() else "no-gil",
    )

    bundled_json_path = os.path.join(
        os.path.abspath(
            os.path.join(os.path.dirname(__file__), "python_internal_offset")
        ),
        "offsets_%s.json" % python_version_id,
    )

    cached_json_path = os.path.join(
        cache_dir,
        "offsets_%s.json" % python_version_id,
    )

    if not isOffsetsJsonOutdated(bundled_json_path, python_version_str):
        return None

    if not isOffsetsJsonOutdated(cached_json_path, python_version_str):
        return None  # Cache already contains mapping

    general.info(
        "Internal structure mapping for Python %s is missing. Extracting offsets dynamically via zig..."
        % python_version_str
    )

    try:
        from nuitka.tools.general.generate_header.GenerateHeader import (
            generateHeader,
        )

        generateHeader()
    except Exception as e:  # pylint: disable=broad-exception-caught
        return general.sysexit("Failed to generate headers dynamically: %s" % str(e))
    return cache_dir


def _getPythonInternalHeadersAndHash(internal_include_dir):
    # Hash the original header contents and mode-dependent adaptation settings.
    hash_obj = Hash()
    header_files = []

    hash_obj.updateFromValues(
        "adapted-python-headers-v1",
        getNuitkaVersion(),
        "module" if shallMakeModule() else "non-module",
    )

    for filename in sorted(getFileList(internal_include_dir, only_suffixes=(".h",))):
        header_files.append(filename)
        hash_obj.updateFromFile(filename)

    return header_files, hash_obj.asHexDigest()


def _shallCreateAdaptedPythonHeaderFiles():
    # Python before 3.13 is not affected.
    if python_version < 0x3D0:
        return False

    needs_windows_mingw_opt_in = isWin32Windows() and not (
        isExperimental("force-mingw64") and isMingw64()
    )
    is_python314_module_mode = shallMakeModule() and 0x3E0 <= python_version < 0x3F0

    if not shallMakeModule():
        uses_default_platform_gate = True
    else:
        # Python 3.14 module mode needs adapted headers on all OSes due to
        # cross-patch interpreter layout changes. For MSVC, offsetof is
        # always correct since the compiler and runtime layouts match.
        uses_default_platform_gate = (
            not is_python314_module_mode or needs_windows_mingw_opt_in
        )

    if uses_default_platform_gate:
        # TODO: We delay this until we have a better way to detect the compiler
        # inside of Nuitka and not in Scons.
        if needs_windows_mingw_opt_in:
            return False

        if not isLinux() and not isWin32Windows():
            return False

    return True


def createAdaptedPythonHeaderFiles(source_dir):
    """Ensure sanitized internal headers exist in the build cache and JSON offsets are extracted."""

    if not _shallCreateAdaptedPythonHeaderFiles():
        return None

    internal_include_dir = os.path.join(
        getTargetPythonIncludePath(
            logger=general,
            python_debug=shallUsePythonDebug(),
            self_compiled_python_uninstalled=isSelfCompiledPythonUninstalled(),
        ),
        "internal",
    )

    header_files, total_hash = _getPythonInternalHeadersAndHash(internal_include_dir)

    cache_base_dir = getCacheDir("adapted_headers", create=True)

    # 1. Ensure Offsets Extracted
    ensurePythonInternalsOffsets(cache_dir=cache_base_dir)

    target_cache_dir = os.path.join(cache_base_dir, total_hash)
    adapt_cache_file = os.path.join(target_cache_dir, "adapted_headers.patch")

    # 2. Check if headers already stripped
    if os.path.exists(adapt_cache_file):
        copyFile(adapt_cache_file, os.path.join(source_dir, "adapted_headers.patch"))

        return target_cache_dir

    # 3. Adapt and Copy Headers
    makePath(os.path.join(target_cache_dir, "internal"))

    adapt_lines = []

    for filename in header_files:
        rel_path = os.path.relpath(filename, internal_include_dir)
        d = os.path.join(target_cache_dir, "internal", rel_path)
        makePath(os.path.dirname(d))

        item = os.path.basename(filename)
        if item.startswith("pycore_"):
            content = getFileContents(filename, mode="r", encoding="utf-8")
            adapted = adaptPythonHeaderFile(content, rel_path.replace(os.path.sep, "/"))
            putTextFileContents(d, adapted, encoding="utf-8")

            if content != adapted:
                # Add unified diff to our patch aggregator
                adapt_lines.extend(
                    getUnifiedDiff(
                        old_lines=content.splitlines(True),
                        new_lines=adapted.splitlines(True),
                        old_filename="a/internal/" + rel_path.replace(os.path.sep, "/"),
                        new_filename="b/internal/" + rel_path.replace(os.path.sep, "/"),
                    )
                )
                if adapt_lines and not adapt_lines[-1].endswith("\n"):
                    adapt_lines.append("\n")
        else:
            copyFile(filename, d)

    if adapt_lines:
        putTextFileContents(adapt_cache_file, "".join(adapt_lines), encoding="utf-8")
        copyFile(adapt_cache_file, os.path.join(source_dir, "adapted_headers.patch"))

    return target_cache_dir


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


from __future__ import print_function

import itertools


def compiled_func(a, b, c, d, e, f):
    return a, b, c, d, e, f


def getUnknownValue():
    return 8


arg_dict = {
    "d": 9,
    "e": 9,
    "f": 9,
}


def calledRepeatedly():
    a = getUnknownValue()
    b = getUnknownValue()
    c = getUnknownValue()

    compiled_f = compiled_func

    # This is supposed to make a call to a compiled function, which is
    # being optimized separately.
    # construct_begin
    compiled_f(a=a, b=b, c=c, **arg_dict)
    compiled_f(a=a, b=b, c=c, **arg_dict)
    compiled_f(a=a, b=b, c=c, **arg_dict)
    # construct_alternative
    pass
    # construct_end

    return compiled_f


for x in itertools.repeat(None, 50000):
    calledRepeatedly()

print("OK.")

#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Tools related to management of copyright headers and footers."""

import ast
import os

from nuitka.tools.quality.ScanSources import isPythonFile
from nuitka.utils.FileOperations import (
    addFileContentsBOM,
    getFileContentByLine,
    getFileContents,
    openTextFile,
    stripFileContentsBOM,
)
from nuitka.Version import getNuitkaVersionYear

copyright_year = getNuitkaVersionYear()
_default_copyright_holder = "Kay Hayen, mailto:kay.hayen@gmail.com"

# spell-checker: ignore Batakrishna,Jorj,Kierzkowski,Pawe,Sahu,Teske

copyright_holder = {
    "nuitka/plugins/standard/TkinterPlugin.py": "Jorj McKie, mailto:<jorj.x.mckie@outlook.de>",
    "nuitka/plugins/standard/NumpyPlugin.py": "Jorj McKie, mailto:<jorj.x.mckie@outlook.de>",
    "nuitka/plugins/standard/GeventPlugin.py": "Jorj McKie, mailto:<jorj.x.mckie@outlook.de>",
    "nuitka/plugins/standard/TensorflowPlugin.py": "Jorj McKie, mailto:<jorj.x.mckie@outlook.de>",
    "nuitka/plugins/standard/SklearnPlugin.py": "Jorj McKie, mailto:<jorj.x.mckie@outlook.de>",
    "nuitka/plugins/standard/PbrPlugin.py": "Jorj McKie, mailto:<jorj.x.mckie@outlook.de>",
    "nuitka/plugins/standard/TorchPlugin.py": "Jorj McKie, mailto:<jorj.x.mckie@outlook.de>",
    "nuitka/plugins/standard/EventletPlugin.py": "Jorj McKie, mailto:<jorj.x.mckie@outlook.de>",
    "nuitka/plugins/standard//OptionsNannyPlugin.py": "Fire-Cube <ben7@gmx.ch>",
    "nuitka/plugins/standard/PlaywrightPlugin.py": "Kevin Rodriguez <mailto:turcioskevinr@gmail.com>",
    "tests/standalone/SocketUsing.py": "Pawe\u0142 Kierzkowski, mailto:<pk.pawelo@gmail.com>",
    "nuitka/nodes/BuiltinAnyNodes.py": "Batakrishna Sahu, mailto:<Batakrishna.Sahu@suiit.ac.in>",
    "nuitka/nodes/BuiltinAllNodes.py": "Batakrishna Sahu, mailto:<Batakrishna.Sahu@suiit.ac.in>",
    "nuitka/nodes/IterationHandles.py": "Batakrishna Sahu, mailto:<Batakrishna.Sahu@suiit.ac.in>",
    "doc/custom.css": "Batakrishna Sahu, mailto:<Batakrishna.Sahu@suiit.ac.in>",
    "doc/py-file.bat": "Jorj McKie, mailto:<jorj.x.mckie@outlook.de>",
    "doc/py-file.sh": "Jorj McKie, mailto:<jorj.x.mckie@outlook.de>",
    "tests/PyPI-pytest/run_all.py": "Tommy Li, mailto:<tommyli3318@gmail.com>",
    "tests/distutils/nested_namespaces/a/b/pkg/__init__.py": "Jan Teske, mailto:<jteske@posteo.net>",
    "tests/distutils/nested_namespaces/setup.py": "Jan Teske, mailto:<jteske@posteo.net>",
    "nuitka/tools/quality/auto_format/YamlFormatter.py": "Fire-Cube <ben7@gmx.ch>",
}

copyright_holder_year = {
    "Batakrishna Sahu, mailto:<Batakrishna.Sahu@suiit.ac.in>": "2020",
    "Jan Teske, mailto:<jteske@posteo.net>": "2020",
    "Jorj McKie, mailto:<jorj.x.mckie@outlook.de>": "2020",
    "Kevin Rodriguez <mailto:turcioskevinr@gmail.com>": "2024",
    "Pawe\u0142 Kierzkowski, mailto:<pk.pawelo@gmail.com>": "2020",
    "Tommy Li, mailto:<tommyli3318@gmail.com>": "2020",
}

# Files that have no suffix, but should be commented anyway, spell-checker: ignore restlint
_plain_files = (
    "Python",
    "compare_with_cpython",
    "compare_with_xml",
    "check-nuitka-with-pylint",
    "check-nuitka-with-pylint3",
    "measure-construct-performance",
    "check-nuitka-with-restlint",
    "check-nuitka-with-yamllint",
    "check-nuitka-with-codespell",
    "check-reference-counts",
    "autoformat-nuitka-source",
    "nuitka-linux-container",
    "generate-specialized-c-code",
    "generate-specialized-python-code",
    "generate-specialized-offsets-code",
    "sort-nuitka-import-statements",
    "find_sxs_modules",
    "find-module",
    "nuitka",
    ".sourcery.yaml",
    "nuitka-watch",
    "nuitka-run",
    "nuitka2",
    "nuitka2-run",
    "nuitka-decrypt",
    "run-tests",
    "run-inside-nuitka-container",
    "custom.css",
    "pre-commit",
    "Containerfile",
    "requirements.txt",
    "requirements-commercial.txt",
    "commercial.nuitka-package.config.yml",
    "runner",
)


def isPlainFileWithCopyright(filename):
    return os.path.basename(filename) in _plain_files


def _normalizeCopyrightFilename(filename):
    return filename.replace("\\", "/")


def getCopyrightHolder(filename):
    return copyright_holder.get(
        _normalizeCopyrightFilename(filename),
        _default_copyright_holder,
    )


def getCopyrightHolderYear(holder):
    return copyright_holder_year.get(holder, str(copyright_year))


def getCopyrightClaim(filename, claim):
    holder = getCopyrightHolder(filename)

    r = [
        (
            "    Copyright %s, %s"
            % (
                getCopyrightHolderYear(holder),
                holder,
            )
        ).encode("utf8"),
        b"",
    ] + claim

    return r


def _formatComments(filename, comments):
    # All cases in one, pylint: disable=too-many-branches

    assert all(type(comment) is bytes for comment in comments), comments
    assert not any(b"\n" in comment for comment in comments), comments

    if not comments:
        return comments

    if filename.endswith(".css"):
        comments.insert(0, b"/" + b"*" * 70)
        comments.append(b"/" + b"*" * 70)
    elif filename.endswith(
        (
            ".py",
            ".sh",
            ".scons",
            ".template",
            ".yml",
            ".cfg",
            ".toml",
            ".ps1",
            ".containerfile",
        )
    ) or isPlainFileWithCopyright(filename):
        comments = [
            (b"# %s" % comment if comment != b"" else b"#") for comment in comments
        ]
    elif filename.endswith(".puml"):
        comments = [
            (b"' %s" % comment if comment != b"" else b"'") for comment in comments
        ]
    elif filename.endswith((".c", ".cpp", ".h", ".S", ".qml")):
        comments = [
            (b"// %s" % comment if comment != b"" else b"//") for comment in comments
        ]
    elif filename.endswith((".cmd", ".bat")):
        comments = [
            (b"rem %s" % comment if comment != b"" else b"rem") for comment in comments
        ]
    elif filename.endswith(".j2"):
        max_len = max(75, max(len(comment.strip()) for comment in comments))
        template = b"{#-  %%-%ds -#}" % max_len
        comments = [(template % comment.strip()) for comment in comments]
    elif (
        filename.endswith(".txt")
        and "tests/commercial" in filename
        or "LICENSE" in filename
    ):
        comments = []
    elif filename.endswith(".rst"):
        comments = []
    elif filename.endswith(".md"):
        comments = [
            (b"<!-- %s -->" % comment if comment != b"" else b"<!-- -->")
            for comment in comments
        ]
    elif filename.endswith(".json"):
        comments = []
    elif filename.endswith(".svg"):
        comments = []
    elif filename.endswith(".png"):
        comments = []
    elif filename.endswith(".zip"):
        comments = []
    else:
        assert False, filename

    return comments


def _isEncodingIndicatorLine(line):
    return (
        line.startswith(b"# -*-")
        or line.startswith(b"# coding:")
        or line.startswith(b"# encoding:")
    )


def _pickLeadingLines(old_lines):
    old_lines_head = []

    if old_lines and (
        old_lines[0].startswith(b"#!") or old_lines[0].startswith(b"@echo")
    ):
        old_lines_head = old_lines[0:1]
        old_lines = old_lines[1:]

    if old_lines and _isEncodingIndicatorLine(old_lines[0]):
        old_lines_head += old_lines[0:1]
        old_lines = old_lines[1:]
    elif (
        not old_lines_head
        and len(old_lines) >= 2
        and old_lines[0].startswith(b"#")
        and _isEncodingIndicatorLine(old_lines[1])
    ):
        old_lines_head = old_lines[0:2]
        old_lines = old_lines[2:]

    return old_lines_head, old_lines


def attachLeadingComment(filename, effective_filename, comments, replacements=()):
    # Many details and complicated algorithm, pylint: disable=too-many-branches,too-many-locals,too-many-statements

    old_contents = getFileContents(filename, mode="rb")
    old_contents, bom = stripFileContentsBOM(old_contents)
    old_lines = old_contents.splitlines()

    if comments and os.path.basename(effective_filename) != os.path.basename(__file__):
        assert b'Part of "Nuitka"' not in b"".join(old_lines), filename

    if comments and comments[1] == b"" and any(line.strip() for line in old_lines):
        pre_comments = comments[:1]
        pre_comments[0] += b" find license text at end of file"
        post_comments = comments[2:]

        while post_comments[-1].strip() in (b"", b"#"):
            del post_comments[-1]
    else:
        pre_comments = comments
        post_comments = []

        if not any(line.strip() for line in old_lines):
            while pre_comments[-1].strip() in (b"", b"#"):
                del pre_comments[-1]

    del comments

    pre_comments = _formatComments(effective_filename, pre_comments)
    post_comments = _formatComments(effective_filename, post_comments)

    is_python = isPythonFile(filename=filename, effective_filename=effective_filename)

    # Do what black will do otherwise
    if pre_comments:
        if is_python:
            pre_comments.append(b"")

    if post_comments:
        post_comments.insert(0, b"")

        if is_python:
            for line in reversed(old_lines):
                if line.startswith((b"def ", b"class ")):
                    bottom_function = True
                    break
                if line.strip() and not chr(line[0]).isalnum():
                    bottom_function = False
                    break
            else:
                bottom_function = False

            if not bottom_function:
                try:
                    last_node = ast.parse(b"\n".join(old_lines)).body[-1]
                except (SyntaxError, IndexError):
                    pass
                else:
                    if type(last_node) is ast.If:
                        # ast API, spell-checker: ignore orelse
                        last_node = (
                            last_node.orelse[-1]
                            if last_node.orelse
                            else last_node.body[-1]
                        )

                    bottom_function = type(last_node) in (
                        ast.FunctionDef,
                        ast.AsyncFunctionDef,
                        ast.ClassDef,
                    )

            if bottom_function:
                post_comments.insert(0, b"")

    old_lines_head, old_lines = _pickLeadingLines(old_lines=old_lines)

    if pre_comments:
        while old_lines and old_lines[0].strip() in (b"", b"#"):
            del old_lines[0]
        old_lines.insert(0, b"")

    lines = old_lines_head + pre_comments + old_lines + post_comments

    if bom:
        lines[0] = addFileContentsBOM(lines[0])

    def replace(line):
        for src, dest in replacements:
            line = line.replace(src, dest)

        return line

    lines = [replace(line) for line in lines]

    if is_python:
        while lines and not lines[-1].strip():
            del lines[-1]

    new_line = b"\r\n" if effective_filename.endswith(".cmd") else b"\n"

    with openTextFile(filename, "wb") as output:
        output.writelines(line + new_line for line in lines)


_copyright_claim_standard = []


def getLicenseTextStandard():
    if not _copyright_claim_standard:
        inside = False

        for line in getFileContentByLine(
            os.path.join(os.path.dirname(__file__), "../../../debian/copyright")
        ):
            if line.startswith("License:"):
                inside = True
                continue

            if not inside:
                continue

            if not line.startswith(" "):
                break

            line = line[1:]

            if line == ".":
                line = ""
            else:
                line = "    " + line

            if "On Debian GNU/Linux systems" in line:
                break

            _copyright_claim_standard.append(line.encode("utf8"))

        while not _copyright_claim_standard[-1].strip():
            del _copyright_claim_standard[-1]

        # Per file, we maintain a different wording.
        _copyright_claim_standard[0] = _copyright_claim_standard[0].replace(
            b"Nuitka is an",
            b'Part of "Nuitka", an',
        )

    return _copyright_claim_standard


_copyright_claim_commercial = (b"""\
    Part of Nuitka commercial.

    Commercial grade license of Nuitka. No distribution to outside of the
    buyer is allowed, the file must be kept secret. Usage is limited to
    Nuitka. Changed versions, or versions of changes sent to the
    copyright holder are automatically licensed to him under the Apache
    License, Version 2.0 as used in proper Nuitka, to allow opening code
    up later.

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
""").split(b"\n")


def getLicenseTextCommercial():
    return _copyright_claim_commercial


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Upload the packaging results and/or push branches."""

import os
import sys

# Unchanged, running from checkout, use the parent directory, the nuitka
# package ought be there.
sys.path.insert(0, os.path.normpath(os.path.join(os.path.dirname(__file__), "..")))

# isort:start

from nuitka.tools.release.Release import checkAtHome, checkBranchName
from nuitka.utils.Execution import check_output

checkAtHome()

nuitka_version = check_output(
    "%s ./bin/nuitka --version" % sys.executable, shell=True
).strip()

branch_name = checkBranchName()

if branch_name == "factory":
    for remote in "origin", "github":
        assert (
            os.system(
                "git push --recurse-submodules=no --force-with-lease %s factory"
                % remote
            )
            == 0
        ), remote

    sys.exit(0)

assert (
    os.system(
        "rsync -rvlpt --exclude=deb_dist dist/ root@ssh.nuitka.net:/var/www/releases/"
    )
    == 0
)

if branch_name.startswith("release") or branch_name == "main":
    pass
elif branch_name == "develop":
    for remote in "origin", "github":
        assert (
            os.system(
                "git push --recurse-submodules=no --tags --force-with-lease %s develop"
                % remote
            )
            == 0
        )
        assert os.system("git push %s main" % remote) == 0, remote
else:
    sys.stdout.write("Skipping for branch '%s'" % branch_name)

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Windows COFF Object generation for payload embedding.

This module provides a pure-Python generator for PE/COFF .obj files
containing a single binary payload representing a data section.
It is used to bypass MSVC's lack of #embed.
"""

import shutil
import struct

from nuitka.utils.FileOperations import getFileSize


def _createCoffStringTable(symbol_bytes):
    string_table_data = b""

    if len(symbol_bytes) <= 8:
        sym_name_field = symbol_bytes.ljust(8, b"\x00")
    else:
        # String table format: 4-byte total size (including size field),
        # followed by null-terminated strings.
        offset = 4 + len(string_table_data)
        string_table_data += symbol_bytes + b"\x00"
        sym_name_field = struct.pack("<II", 0, offset)

    if not string_table_data:
        string_table = b""
    else:
        string_table_size = 4 + len(string_table_data)
        string_table = struct.pack("<I", string_table_size) + string_table_data

    return sym_name_field, string_table


def generateWindowsCoffObject(
    in_filename, out_filename, symbol_name, architecture, writeable
):
    """Generate a valid MSVC COFF .obj file containing the given payload.

    Args:
        in_filename (str): Path to the input binary payload.
        out_filename (str): Path to write the resulting .obj file.
        symbol_name (str): The C symbol name to export.
        architecture (str): 'x86', 'x86_64', or 'arm64'.
        writeable (bool): Use a writable '.data' section instead of '.rdata'.
    """

    # pylint: disable=too-many-locals
    arch_map = {
        "x86": 0x014C,  # IMAGE_FILE_MACHINE_I386
        "x86_64": 0x8664,  # IMAGE_FILE_MACHINE_AMD64
        "arm64": 0xAA64,  # IMAGE_FILE_MACHINE_ARM64
    }

    machine = arch_map[architecture]

    # For 32-bit x86, MSVC conventions require a leading underscore on C symbols.
    if architecture == "x86" and not symbol_name.startswith("_"):
        symbol_name = "_" + symbol_name

    symbol_bytes = symbol_name
    if str is not bytes:
        symbol_bytes = symbol_name.encode("utf8")

    sym_name_field, string_table = _createCoffStringTable(symbol_bytes)

    payload_size = getFileSize(in_filename)
    section_data_size = payload_size

    # 4-byte alignment padding for the symbol table
    padding_len = (4 - (section_data_size % 4)) % 4
    padding = b"\x00" * padding_len

    # Header sizes
    file_header_size = 20
    section_header_size = 40

    pointer_to_raw_data = file_header_size + section_header_size
    pointer_to_symbol_table = pointer_to_raw_data + section_data_size + padding_len

    # 1. PE File Header
    file_header = struct.pack(
        "<HHIIIHH",  # spell-checker: ignore HHIIIHH
        machine,
        1,  # NumberOfSections
        0,  # TimeDateStamp (0 for reproducibility)
        pointer_to_symbol_table,
        1,  # NumberOfSymbols
        0,  # SizeOfOptionalHeader
        0,  # Characteristics
    )

    # 2. Section Header
    if writeable:
        section_name = b".data\x00\x00\x00"

        # Characteristics flags:
        # IMAGE_SCN_CNT_INITIALIZED_DATA = 0x00000040
        # IMAGE_SCN_ALIGN_16BYTES        = 0x00500000
        # IMAGE_SCN_MEM_READ             = 0x40000000
        # IMAGE_SCN_MEM_WRITE            = 0x80000000
        section_characteristics = 0xC0500040
    else:
        section_name = b".rdata\x00\x00"  # spell-checker: ignore rdata

        # Characteristics flags:
        # IMAGE_SCN_CNT_INITIALIZED_DATA = 0x00000040
        # IMAGE_SCN_ALIGN_16BYTES        = 0x00500000
        # IMAGE_SCN_MEM_READ             = 0x40000000
        section_characteristics = 0x40500040

    section_header = struct.pack(
        "<8sIIIIIIHHI",  # spell-checker: ignore IIIIIIHHI
        section_name,
        0,  # VirtualSize (0 for obj)
        0,  # VirtualAddress (0 for obj)
        section_data_size,
        pointer_to_raw_data,
        0,  # PointerToRelocations
        0,  # PointerToLinenumbers
        0,  # NumberOfRelocations
        0,  # NumberOfLinenumbers
        section_characteristics,
    )

    # 3. Symbol Table Entry
    # Value is 0 (offset within section)
    # SectionNumber is 1 (1-based index to our section)
    # Type is 0 (IMAGE_SYM_TYPE_NULL)
    # StorageClass is 2 (IMAGE_SYM_CLASS_EXTERNAL)
    # NumberOfAuxSymbols is 0
    symbol_table_entry = struct.pack(
        "<8sIhHBB",
        sym_name_field,
        0,  # Value
        1,  # SectionNumber
        0,  # Type
        2,  # StorageClass
        0,  # NumberOfAuxSymbols
    )

    with open(out_filename, "wb") as f_out:
        f_out.write(file_header)
        f_out.write(section_header)

        with open(in_filename, "rb") as f_in:
            shutil.copyfileobj(f_in, f_out)

        f_out.write(padding)
        f_out.write(symbol_table_entry)

        if string_table:
            f_out.write(string_table)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Signing of executables."""

import re

from nuitka.options.Options import (
    getMacOSSignedAppName,
    getMacOSSigningCertificateFilename,
    getMacOSSigningCertificatePassword,
    getMacOSSigningIdentity,
    shallUseSigningForNotarization,
)
from nuitka.Tracing import postprocessing_logger

from .Execution import executeToolChecked
from .FileOperations import getExternalUsePath, withMadeWritableFileMode

_macos_codesign_usage = (
    "The 'codesign' is used to make signatures on macOS and required to be found."
)

_macos_security_usage = (
    "The 'security' is used to access signatures from files on macOS."
)

_macos_find_identity_pattern = re.compile(
    r'^\s*\d+\)\s+([0-9A-F]+)\s+"(.*)"(?:\s+\(.*\))?$'
)


def _filterCodesignErrorOutput(stderr, signing_identity, signing_keychain_filename):
    stderr = b"\n".join(
        line
        for line in stderr.splitlines()
        if line
        if b"replacing existing signature" not in line
        if b"unable to build chain to self-signed root" not in line
    )

    if b"errSecInternalComponent" in stderr:
        details = []

        if signing_identity is not None:
            details.append("identity '%s'" % signing_identity)

        if signing_keychain_filename is not None:
            details.append("keychain '%s'" % signing_keychain_filename)

        if details:
            details = " Requested %s." % " and ".join(details)
        else:
            details = ""

        return postprocessing_logger.sysexit("""\
The macOS codesign step failed with a keychain access error \
'errSecInternalComponent'. This can mean that the signing identity is \
missing, incomplete, locked, or not accessible to codesign. Check that the \
requested identity exists with its private key and that the relevant keychain \
can be unlocked and used for code signing.%s""" % details)

    return None, stderr


def _filterSecurityErrorOutput(stderr):
    # TODO: Probably some errors are given if it already exists.
    return None, stderr


def _getMacOSCodeSigningIdentities(signing_keychain_filename):
    command = ["security", "find-identity", "-p", "codesigning"]

    if signing_keychain_filename is None:
        command.append("-v")

    if signing_keychain_filename is not None:
        command.append(signing_keychain_filename)

    output = executeToolChecked(
        logger=postprocessing_logger,
        command=command,
        absence_message="The 'security' program is used to scan for signing identities",
        stderr_filter=_filterSecurityErrorOutput,
        decoding=str is not bytes,
    )

    result = []
    seen_identities = set()

    for line in output.splitlines():
        match = _macos_find_identity_pattern.match(line)

        if match is None:
            continue

        identity_hash, signing_name = match.groups()

        if identity_hash in seen_identities:
            continue

        seen_identities.add(identity_hash)
        result.append((identity_hash, signing_name))

    return tuple(result)


def _unlockMacIdentityKeychain(signing_keychain_filename):
    command = [
        "security",
        "unlock-keychain",
        "-p",
        getMacOSSigningCertificatePassword() or "",
        signing_keychain_filename,
    ]

    executeToolChecked(
        logger=postprocessing_logger,
        command=command,
        absence_message=_macos_security_usage,
        stderr_filter=_filterSecurityErrorOutput,
    )


def _detectMacOSCodeSigningIdentity(signing_keychain_filename=None):
    identities = _getMacOSCodeSigningIdentities(
        signing_keychain_filename=signing_keychain_filename
    )

    if len(identities) > 1:
        found_names = ", ".join(
            "'%s'" % signing_name for _identity_hash, signing_name in identities
        )

        if signing_keychain_filename is None:
            return postprocessing_logger.sysexit(
                "More than one signing identity, auto mode cannot be used. Found: %s."
                % found_names
            )
        else:
            return postprocessing_logger.sysexit(
                "More than one code signing identity in keychain '%s', auto mode cannot be used. Found: %s."
                % (signing_keychain_filename, found_names)
            )

    if not identities:
        if signing_keychain_filename is None:
            return postprocessing_logger.sysexit(
                "Failed to detect any signing identity, auto mode cannot be used."
            )
        else:
            return postprocessing_logger.sysexit(
                "Failed to detect any code signing identity in keychain '%s', auto mode cannot be used."
                % signing_keychain_filename
            )

    identity_hash, signing_name = identities[0]

    if signing_keychain_filename is None:
        postprocessing_logger.info(
            "Using signing identity %s automatically." % signing_name
        )
    else:
        postprocessing_logger.info(
            "Using signing identity %s from keychain '%s' automatically."
            % (signing_name, signing_keychain_filename)
        )

    if signing_keychain_filename is not None:
        return signing_name
    else:
        return identity_hash


def addMacOSCodeSignature(filenames, entitlements_filename):
    """Add the code signature to filenames.

    Args:
        filenames - The filenames to be signed.
        entitlements_filename - The entitlements file to use, optional.

    Returns:
        None

    Notes:
        This is macOS specific.
    """

    # Weak signing.
    identity = getMacOSSigningIdentity()

    signing_keychain_filename = getMacOSSigningCertificateFilename()

    if signing_keychain_filename is not None:
        signing_keychain_filename = getExternalUsePath(signing_keychain_filename)
        _unlockMacIdentityKeychain(signing_keychain_filename)

    if identity == "auto":
        identity = _detectMacOSCodeSigningIdentity(
            signing_keychain_filename=signing_keychain_filename
        )

    command = [
        # Need to avoid Anaconda codesign.
        "/usr/bin/codesign",
        "-s",
        identity,
        "--force",
        "--deep",
        "--preserve-metadata=entitlements",
    ]

    if signing_keychain_filename is not None:
        command += [
            "--keychain",
            signing_keychain_filename,
        ]

    macos_signed_app_name = getMacOSSignedAppName()

    if macos_signed_app_name is not None:
        command += [
            "-i",
            macos_signed_app_name,
        ]

    if entitlements_filename is not None:
        command += [
            "--entitlements",
            getExternalUsePath(entitlements_filename),
        ]

    if shallUseSigningForNotarization():
        command.append("--options=runtime")

    assert type(filenames) is not str
    command.extend(filenames)

    with withMadeWritableFileMode(filenames):
        executeToolChecked(
            logger=postprocessing_logger,
            command=command,
            absence_message=_macos_codesign_usage,
            stderr_filter=_filterCodesignErrorOutput,
            context={
                "signing_identity": identity,
                "signing_keychain_filename": signing_keychain_filename,
            },
        )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/env python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Runner for program tests of Nuitka.

Program tests are typically aiming at checking specific module constellations
and making sure the details are being right there. These are synthetic small
programs, each of which try to demonstrate one or more points or special
behavior.

"""

import os
import sys

# Find nuitka package relative to us.
sys.path.insert(
    0,
    os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..")
    ),
)

# isort:start

from nuitka.containers.OrderedDicts import OrderedDict
from nuitka.reports.CompilationReportReader import (
    extractModulesUsedByModule,
    parseCompilationReport,
)
from nuitka.tools.testing.Common import (
    checkTestRequirements,
    compareWithCPython,
    createSearchMode,
    my_print,
    reportSkip,
    scanDirectoryForTestCaseFolders,
    setup,
    test_logger,
    withPythonPathChange,
)
from nuitka.utils.Execution import check_call
from nuitka.utils.FileOperations import (
    makePath,
    putTextFileContents,
    resetDirectory,
)


def checkMultiprocessingUsingReport(filename, python_version):
    compilation_report = parseCompilationReport("compilation-report-%s.xml" % filename)

    parent_main_module = None

    for module_node in compilation_report.findall("module"):
        if module_node.attrib["name"] == "__parents_main__":
            parent_main_module = module_node
            break

    assert parent_main_module is not None, [
        module_node.attrib["name"]
        for module_node in compilation_report.findall("module")
    ]
    assert (
        parent_main_module.attrib["reason"]
        == "Auto enable multiprocessing/anyio freeze support"
    ), parent_main_module.attrib

    # On Python 2, the fake module is executed directly for freeze support.
    # On Python 3, the multiprocessing post-load helper imports it.
    if python_version < (3,):
        return

    modules_used = extractModulesUsedByModule(
        compilation_report=compilation_report,
        module_name="multiprocessing-postLoad",
    )

    assert modules_used["__parents_main__"]["finding"] == "fake", modules_used


def _makeBytecodeOnlyModule(filename, source_code):
    putTextFileContents(filename=filename, contents=source_code, encoding="utf8")

    try:
        check_call(
            (
                os.environ["PYTHON"],
                "-c",
                # spell-checker: ignore cfile,doraise
                "import py_compile,sys; py_compile.compile(sys.argv[1], cfile=sys.argv[1] + 'c', doraise=True)",
                filename,
            )
        )
    finally:
        if os.path.exists(filename):
            os.unlink(filename)


def updateBytecodeOnlyImportPath(test_directory):
    bytecode_only_path = os.path.join(test_directory, "path1")

    resetDirectory(
        path=bytecode_only_path,
        logger=test_logger,
        ignore_errors=True,
        extra_recommendation=None,
    )

    _makeBytecodeOnlyModule(
        filename=os.path.join(bytecode_only_path, "bytecode_only_module.py"),
        source_code='module_value = "module-bytecode-only"\n',
    )

    package_dir = os.path.join(bytecode_only_path, "bytecode_only_package")
    makePath(package_dir)

    _makeBytecodeOnlyModule(
        filename=os.path.join(package_dir, "__init__.py"),
        source_code='package_value = "package-bytecode-only"\n',
    )

    _makeBytecodeOnlyModule(
        filename=os.path.join(package_dir, "submodule.py"),
        source_code='submodule_value = "submodule-bytecode-only"\n',
    )


def main():
    # Complex stuff, even more should become common code though.
    # pylint: disable=too-many-branches,too-many-statements

    python_version = setup(suite="programs", needs_io_encoding=True)

    search_mode = createSearchMode()

    extra_options = os.getenv("NUITKA_EXTRA_OPTIONS", "")

    for filename, filename_main in scanDirectoryForTestCaseFolders("."):
        active = search_mode.consider(dirname=None, filename=filename)

        if not active:
            continue

        # For these, we expect that they will fail.
        expected_errors = [
            "module_exits",
            "main_raises",
            "main_raises2",
            "package_contains_main",
        ]

        # After Python3 those have been made to work.
        if python_version < (3, 5):
            expected_errors.append("cyclic_imports")

        # Allowed with Python3, packages need no more "__init__.py"
        if python_version < (3,):
            expected_errors.append("package_missing_init")

        # Allowed with Python3.5 only:
        if python_version < (3, 5):
            expected_errors.append("package_init_issue")

        # Allowed with Python3, name imports can be module imports
        if python_version < (3,):
            expected_errors.append("named_imports")

        extra_variant = []

        if filename not in expected_errors:
            extra_flags = ["expect_success"]
        else:
            extra_flags = ["expect_failure"]

        # We annotate some tests, use that to lower warnings.
        extra_flags.append("plugin_enable:pylint-warnings")

        if filename in (
            "reimport_main_static",
            "package_missing_init",
            "dash_import",
            "package_contains_main",
            "case_imports3",
            "import_variants",
            "package_init_import",
            "pkgutil_iter_modules",
        ):
            extra_flags.append("ignore_warnings")

        extra_flags.append("remove_output")

        extra_flags.append("--follow-imports")

        # Run it as a package and also as directory.
        if filename == "package_program":
            # Not really supported for 2.6
            if python_version >= (2, 7):
                extra_variant.append("--python-flag=-m")

        # Cannot include the files with syntax errors, these would then become
        # ImportError, but that's not the test. In all other cases, use two
        # step execution, which will not add the program original source to
        # PYTHONPATH.
        if filename != "syntax_errors":
            extra_flags.append("two_step_execution")
        else:
            extra_flags.append("binary_python_path")

        if filename == "plugin_import":
            os.environ["NUITKA_EXTRA_OPTIONS"] = (
                extra_options + " --include-package=some_package"
            )
        elif filename == "reimport_main_dynamic":
            if python_version < (3,):
                os.environ["NUITKA_EXTRA_OPTIONS"] = (
                    extra_options
                    + " --include-plugin-directory=%s" % (os.path.abspath(filename))
                )
            else:
                os.environ["NUITKA_EXTRA_OPTIONS"] = (
                    extra_options
                    + " --include-plugin-files=%s/*.py" % (os.path.abspath(filename))
                )

            extra_flags.append("ignore_warnings")
        elif filename == "multiprocessing_using":
            extra_flags.append("--report=compilation-report-%s.xml" % filename)
        else:
            os.environ["NUITKA_EXTRA_OPTIONS"] = extra_options

        requirements_met, error_message = checkTestRequirements(
            os.path.join(filename, filename_main)
        )
        if not requirements_met:
            reportSkip(error_message, ".", filename)
            continue

        my_print("Consider output of recursively compiled program:", filename)

        if filename == "bytecode_only_import":
            updateBytecodeOnlyImportPath(filename)

        extra_python_path = [
            os.path.abspath(os.path.join(filename, entry))
            for entry in os.listdir(filename)
            if entry.startswith("path")
        ]

        if extra_python_path:
            my_print("Applying extra PYTHONPATH %r." % extra_python_path)

        with withPythonPathChange(extra_python_path):
            test_variants = OrderedDict((("", extra_flags),))

            if extra_variant:
                test_variants["variant"] = extra_flags + extra_variant

            compareWithCPython(
                dirname=filename,
                filename=filename_main,
                extra_flags=test_variants,
                search_mode=search_mode,
            )

            if filename == "multiprocessing_using":
                checkMultiprocessingUsingReport(filename, python_version)

    search_mode.finish()


if __name__ == "__main__":
    main()

#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""OpenSUSE Build Service (OSC) upload release tool.

Uploads Nuitka branches adapting the RPM configuration to the different
projects on OSC.
"""

import os
import shutil
import sys

from nuitka.tools.Basics import goHome
from nuitka.tools.release.Release import (
    checkBranchName,
    cleanupSourceDistributionState,
)
from nuitka.Tracing import tools_logger
from nuitka.utils.FileOperations import (
    copyFile,
    getFileContents,
    makePath,
    putTextFileContents,
    renameFile,
    withDirectoryChange,
)
from nuitka.Version import getNuitkaVersion


def main():
    goHome()

    nuitka_version = getNuitkaVersion()
    branch_name = checkBranchName()

    shutil.rmtree("build", ignore_errors=True)
    makePath("build")

    cleanupSourceDistributionState()

    # Used by rpmbuild, spell-checker: ignore rpmbuild
    makePath(os.path.expanduser("~/rpmbuild/SOURCES"))

    # Upload stable releases to OpenSUSE Build Service:
    if (
        branch_name.startswith("release")
        or branch_name.startswith("hotfix")
        or branch_name == "main"
    ):
        rpm_project_name = "Nuitka"
        spec_suffix = ""
    elif branch_name == "develop":
        rpm_project_name = "Nuitka-Develop"
        spec_suffix = "-develop"
    elif branch_name == "factory":
        rpm_project_name = "Nuitka-Factory"
        spec_suffix = "-factory"
    else:
        return tools_logger.sysexit("Skipping RPM build for branch '%s'" % branch_name)

    with withDirectoryChange("build"):
        tools_logger.info(
            "Building source distribution for %s %s"
            % (rpm_project_name, nuitka_version)
        )
        # spell-checker: ignore gztar,rpmlintrc
        assert os.system("%s ../setup.py sdist --formats=gztar" % sys.executable) == 0
        nuitka_dist_filename = "Nuitka%s-%s.tar.gz" % (spec_suffix, nuitka_version)
        renameFile(
            source_filename=os.path.join("..", "dist", nuitka_dist_filename),
            dest_filename=os.path.join("/root/rpmbuild/SOURCES", nuitka_dist_filename),
        )

        copyFile(
            source_path=os.path.join(os.path.dirname(__file__), "nuitka-rpmlintrc"),
            dest_path="/root/rpmbuild/SOURCES/nuitka-rpmlintrc",
        )

        # Adapt the spec file dynamically to version and project name of Nuitka being built.
        spec_contents = getFileContents(
            os.path.join(os.path.dirname(__file__), "nuitka.spec")
        )
        spec_contents = spec_contents.replace("PROJECT_NAME", "nuitka" + spec_suffix)
        spec_contents = spec_contents.replace("PROJECT_VERSION", nuitka_version)

        spec_filename = "nuitka%s.spec" % spec_suffix
        putTextFileContents(
            filename=spec_filename,
            contents=spec_contents,
        )

        os.system("rpmbuild --target x86_64 -bb %s" % spec_filename)


if __name__ == "__main__":
    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#!/usr/bin/python
#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Release tool to sync Developer Manual with code comments."""

import inspect
import re
import sys

from nuitka.utils.FileOperations import getFileContentByLine


def main():
    quote_start_re = re.compile("[Qq]uoting the ``(.*)`` documentation")
    quote_end_re = re.compile("(End|end) quoting the ``(.*)`` documentation")

    quoting = False

    for line in getFileContentByLine("Developer_Manual.rst"):
        if not quoting:
            print(line, end="")

        if not quoting:
            match = quote_start_re.search(line)

            if match:
                quoting = match.group(1)

                if "." in quoting:
                    import_from, import_value = quoting.rsplit(".", 1)

                    # Hopefully OK for us, pylint: disable=W0122
                    exec("from %s import %s" % (import_from, import_value))
                    item = getattr(sys.modules[import_from], import_value)

                    # Should potentially be derived from quoting line.
                    indentation = " " * line.find("Quoting")

                    # Empty line to separate
                    print()

                    for quote_line in inspect.getdoc(item).splitlines():
                        if quote_line:
                            print(indentation + quote_line)
                        else:
                            print()

                    print()
                else:
                    assert False, quoting

        if quoting:
            match = quote_end_re.search(line)

            if match:
                assert quoting == match.group(1)
                quoting = False

                print(line, end="")

    if quoting:
        sys.exit("Error, waiting for end of quote for %s failed" % quoting)


if __name__ == "__main__":
    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


from package1.module1 import module1_f1
from package1.module2 import module2_f1
from package1.subpackage1.submodule11 import submodule11_f2
from package1.subpackage2.submodule21 import submodule21_f1


def main():
    module1_f1("package1.main.py")
    module2_f1("package1.main.py")
    submodule11_f2("package1.main.py")
    submodule21_f1("package1.main.py")


#     Python test originally created or extracted from other peoples work. The
#     parts from me are licensed as below. It is at least Free Software where
#     it's copied from other people. In these cases, that will normally be
#     indicated.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Generation of Nuitka documentation."""

import os
import sys

from nuitka.Tracing import my_print
from nuitka.utils.Execution import check_call, withEnvironmentVarOverridden
from nuitka.utils.FileOperations import (
    getFileContents,
    getFileList,
    openTextFile,
    putTextFileContents,
    replaceFileAtomic,
)


def _optimizePNGs(pngList):
    for png in pngList:
        check_call(["optipng", "-quiet", "-o2", "%s.png" % png])


def makeLogoImages():
    basePathLogo = "doc/Logo/Nuitka-Logo-%s"

    for logo in ("Vertical", "Symbol", "Horizontal"):
        cmd = "convert -background none %s.svg %s.png" % (basePathLogo, basePathLogo)
        check_call((cmd % (logo, logo)).split())

    _optimizePNGs(
        [basePathLogo % item for item in ("Vertical", "Symbol", "Horizontal")]
    )

    if os.path.exists("../Nuitka-website"):
        cmd = "convert -resize %s doc/Logo/Nuitka-Logo-Symbol.svg %s"
        for icon, size in {
            "../Nuitka-website/files/favicon.ico": "32x32",
            "../Nuitka-website/files/favicon.png": "32x32",
            "../Nuitka-website/doc/_static/favicon.ico": "32x32",
            "../Nuitka-website/doc/_static/favicon.png": "32x32",
            "../Nuitka-website/doc/_static/apple-touch-icon-ipad.png": "72x72",
            "../Nuitka-website/doc/_static/apple-touch-icon-ipad3.png": "144x144",
            "../Nuitka-website/doc/_static/apple-touch-icon-iphone.png": "57x57",
            "../Nuitka-website/doc/_static/apple-touch-icon-iphone4.png": "114x114",
            "../Nuitka-website/doc/_static/apple-touch-icon-180x180.png": "180x180",
        }.items():
            check_call((cmd % (icon, size)).split())


# spell-checker: ignore asciinema,postlist,toctree,automodule
extra_rst_keywords = (
    b"asciinema",
    b"postlist",
    b"post",
    b"youtube",
    b"grid",
    b"toctree",
    b"automodule",
    b"dropdown",
    b"rst-class",
    b"carousel",
    b"carousel-content",
    b"carousel-container",
    b"carousel-side-tab",
)
extra_rst_keywords = tuple(sorted(extra_rst_keywords, key=len, reverse=True))


def checkRstLint(document):
    contents = getFileContents(document, mode="rb")

    for keyword in extra_rst_keywords:
        contents = contents.replace(b".. %s::" % keyword, b".. raw:: %s" % keyword)

    import restructuredtext_lint  # pylint: disable=I0021,import-error

    my_print("Checking '%s' for proper restructured text ..." % document, style="blue")
    lint_results = restructuredtext_lint.lint(
        contents.decode("utf8"),
        document,
    )

    lint_error = False
    for lint_result in lint_results:
        # Not an issue.
        if lint_result.message.startswith("Duplicate implicit target name:"):
            continue

        # We switched to raw, but attributes will still bne unknown.
        if lint_result.message.startswith(
            'Error in "raw" directive:\nunknown option: "hidden"'
        ):
            continue
        if lint_result.message.startswith(
            'Error in "raw" directive:\nunknown option: "excerpts"'
        ):
            continue
        if lint_result.message.startswith(
            'Error in "raw" directive:\nunknown option: "members"'
        ):
            continue

        my_print(lint_result, style="yellow")
        lint_error = True

    if lint_error:
        sys.exit("Error, no lint clean rest.")

    my_print("OK.", style="blue")


def _fixupManPageContents(manpage_filename):
    manpage_contents = getFileContents(manpage_filename).splitlines()

    for month in (
        "January",
        "February",
        "March",
        "April",
        "May",
        "June",
        "July",
        "August",
        "September",
        "October",
        "November",
        "December",
    ):
        manpage_contents[1] = manpage_contents[1].replace(month + " ", "")
    manpage_contents[1] = manpage_contents[1].replace("rc0", "")

    new_contents = []
    mark = False

    for count, line in enumerate(manpage_contents):
        if line.startswith(
            (
                "Python:",
                "Commercial:",
                "Flavor:",
                "Executable:",
                "OS:",
                "Arch:",
                "Distribution:",
                "Version C compiler:",
            )
        ):
            continue

        if mark:
            line = ".SS " + line + ".BR\n"
            mark = False
        elif line == ".IP\n" and manpage_contents[count + 1].endswith(":\n"):
            mark = True
            continue

        if line == r"\fB\-\-g\fR++\-only" + "\n":
            line = r"\fB\-\-g\++\-only\fR" + "\n"

        new_contents.append(line)

    putTextFileContents(manpage_filename, contents=new_contents)


def updateManPages():
    with withEnvironmentVarOverridden("NUITKA_MANPAGE_GEN", "1"):
        if not os.path.exists("man"):
            os.mkdir("man")

        cmd = [
            "help2man",
            "-n",
            "the Python compiler",
            "--no-discard-stderr",
            "--no-info",
            "--include",
            "doc/nuitka-man-include.txt",
            "%s ./bin/nuitka --help-plugins" % sys.executable,
        ]

        with openTextFile(os.path.join("doc", "nuitka.1.tmp"), "wb") as output:
            check_call(cmd, stdout=output)
        _fixupManPageContents("doc/nuitka.1.tmp")
        replaceFileAtomic("doc/nuitka.1.tmp", "doc/nuitka.1")

        cmd[-1] += " -run"
        with openTextFile(os.path.join("doc", "nuitka-run.1.tmp"), "wb") as output:
            check_call(cmd, stdout=output)
        _fixupManPageContents("doc/nuitka-run.1.tmp")
        replaceFileAtomic("doc/nuitka-run.1.tmp", "doc/nuitka-run.1")


def checkReleaseDocumentation():
    documents = [
        entry
        for entry in getFileList(".")
        if entry.endswith(".rst") and not entry.startswith("web" + os.path.sep)
        if "inline_copy" not in entry
    ]

    for document in ("README.rst", "Developer_Manual.rst", "Changelog.rst"):
        assert document in documents, documents

    for document in documents:
        checkRstLint(document)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Utility module for generating diffs.

This wraps difflib to keep all diffing logic centralized and isolated.
"""

# spell-checker: ignore lineterm,tofile,fromlines,tolines,fromdesc,todesc,numlines

import difflib
import os

from .FileOperations import getFileContents, listDir


def getUnifiedDiff(
    old_lines, new_lines, old_filename, new_filename, num_lines=3, lineterm="\n"
):
    """Generate a unified diff from sequences of lines."""
    return difflib.unified_diff(
        a=old_lines,
        b=new_lines,
        fromfile=old_filename,
        tofile=new_filename,
        n=num_lines,
        lineterm=lineterm,
    )


def getHtmlDiffTable(
    old_lines, new_lines, old_desc, new_desc, context=False, num_lines=5
):
    """Generate an HTML diff table from sequences of lines."""
    return difflib.HtmlDiff().make_table(
        fromlines=old_lines,
        tolines=new_lines,
        fromdesc=old_desc,
        todesc=new_desc,
        context=context,
        numlines=num_lines,
    )


def compareDirectories(
    logger,
    my_print,
    dir1,
    dir2,
    ignore_suffixes=None,
    ignore_substrings=None,
):
    """Compare directories recursively."""
    # Many details and cases to deal with
    # pylint: disable=too-many-branches,too-many-locals,too-many-statements

    if ignore_suffixes is None:
        ignore_suffixes = ()
    else:
        ignore_suffixes = tuple(ignore_suffixes)

    if ignore_substrings is None:
        ignore_substrings = ()
    else:
        ignore_substrings = tuple(ignore_substrings)

    done = set()
    result = False

    for path1, filename in listDir(dir1):
        if any(substr in path1 for substr in ignore_substrings):
            continue

        path2 = os.path.join(dir2, filename)
        done.add(path1)

        if filename.endswith(ignore_suffixes):
            continue

        if any(substr in filename for substr in ignore_substrings):
            continue

        if not os.path.exists(path2):
            logger.warning("Only in %s: %s" % (dir1, filename))
            result = False
            continue

        if os.path.isdir(path1):
            r = compareDirectories(
                logger=logger,
                my_print=my_print,
                dir1=path1,
                dir2=path2,
                ignore_suffixes=ignore_suffixes,
                ignore_substrings=ignore_substrings,
            )
            if r:
                result = True
        elif os.path.isfile(path1):
            if str is bytes:
                content1 = getFileContents(path1, mode="rb")
                content2 = getFileContents(path2, mode="rb")
            else:
                content1 = getFileContents(path1, encoding="latin1")
                content2 = getFileContents(path2, encoding="latin1")

            diff = getUnifiedDiff(
                old_lines=content1.splitlines(),
                new_lines=content2.splitlines(),
                old_filename=path1,
                new_filename=path2,
                num_lines=3,
            )

            diff_list = list(diff)

            if diff_list:
                for line in diff_list:
                    try:
                        my_print(line)
                    except UnicodeEncodeError:
                        my_print(repr(line))

                result = True
        else:
            assert False, path1

    for path2, filename in listDir(dir2):
        if any(substr in path2 for substr in ignore_substrings):
            continue

        path1 = os.path.join(dir1, filename)

        if path1 in done:
            continue

        if not os.path.exists(path1):
            logger.warning("Only in %s: %s" % (dir2, filename))
            result = False
            continue

    return result


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2026, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Release related common functionality."""

import os
import shutil

from nuitka.tools.Basics import getHomePath
from nuitka.tools.environments.Virtualenv import withVirtualenv
from nuitka.Tracing import tools_logger
from nuitka.utils.Execution import NuitkaCalledProcessError, check_output
from nuitka.utils.FileOperations import (
    deleteFile,
    getFileContents,
    getFileFirstLine,
    getNormalizedPath,
    listDir,
    removeDirectory,
    withDirectoryChange,
)
from nuitka.utils.InstalledPythons import findInstalledPython
from nuitka.utils.Utils import isLinux
from nuitka.Version import getNuitkaVersion, getNuitkaVersionTuple


def getGitDir():
    git_dir = getNormalizedPath(
        os.path.join(os.path.dirname(__file__), "..", "..", "..", ".git")
    )

    if os.path.isdir(git_dir):
        return git_dir

    line = getFileFirstLine(git_dir, "r").strip()
    return getNormalizedPath(
        os.path.join(os.path.dirname(__file__), "..", "..", "..", line[8:])
    )


def checkAtHome(expected="Nuitka Staging"):
    assert os.path.isfile("setup.py")

    git_dir = getGitDir()

    git_description_filename = os.path.join(git_dir, "description")
    description = getFileContents(git_description_filename).strip()

    assert description == expected, (expected, description)


def _getGitCommandOutput(command):
    if type(command) is str:
        command = command.split()

    home_path = getHomePath()

    with withDirectoryChange(home_path):
        output = check_output(command).strip()

    if str is not bytes:
        output = output.decode()

    return output


def getBranchName():
    """Get the git branch name currently running from."""
    # Using a fallback for old git, hopefully not necessary as much anymore.
    try:
        return _getGitCommandOutput("git branch --show-current")
    except NuitkaCalledProcessError:
        return _getGitCommandOutput("git symbolic-ref --short HEAD")


def getBranchRemoteName():
    """Get the git remote name of the branch currently running from."""
    return _getGitCommandOutput("git config branch.%s.remote" % getBranchName())


def getBranchRemoteUrl():
    """Get the git remote url of the branch currently running from."""
    return _getGitCommandOutput("git config remote.%s.url" % getBranchRemoteName())


def getBranchRemoteIdentifier():
    """Get the git remote identifier of the branch currently running from.

    This identifier is used to classify git origins, they might be github,
    private git, or unknown.
    """

    branch_remote_url = getBranchRemoteUrl()

    branch_remote_host = branch_remote_url.split(":", 1)[0].split("@")[-1]

    if branch_remote_host.endswith(".home"):
        branch_remote_host = branch_remote_host.rsplit(".", 1)[0]

    if branch_remote_host == "git-server":
        return "internal"
    elif branch_remote_host.endswith("nuitka.net"):
        return "private"
    elif branch_remote_host == "github":
        return "public"
    else:
        return "unknown"


def checkBranchName():
    branch_name = getBranchName()

    nuitka_version = getNuitkaVersion()

    assert branch_name in (
        "main",
        "develop",
        "factory",
        "release/" + nuitka_version,
        "hotfix/" + nuitka_version,
    ), branch_name

    return branch_name


def getBranchCategory(branch_name):
    """There are 3 categories of releases. Map branch name on them."""

    if (
        branch_name.startswith("release")
        or branch_name == "main"
        or branch_name.startswith("hotfix/")
    ):
        category = "stable"
    elif branch_name == "factory":
        category = "factory"
    elif branch_name == "develop":
        category = "develop"
    else:
        assert False

    return category


def _makeDistFilenameBase(nuitka_version):
    if nuitka_version[3] is False:
        return "dist/nuitka-%d.%drc%d" % (
            nuitka_version[0],
            nuitka_version[1],
            nuitka_version[-1],
        )
    else:
        if nuitka_version[2] == 0:
            return "dist/nuitka-%d.%d" % (
                nuitka_version[0],
                nuitka_version[1],
            )
        else:
            return "dist/nuitka-%d.%d.%d" % nuitka_version[:3]


def cleanupSourceDistributionState():
    """Remove cached source distribution metadata from the checkout.

    Persistent release worker checkouts can keep an outdated 'MANIFEST' or
    project '.egg-info' directory around. That may cause 'setup.py sdist' to
    omit newly added non-code files from the next source archive.
    """

    deleteFile("MANIFEST", must_exist=False)

    for fullpath, filename in listDir("."):
        if filename.lower().endswith(".egg-info") and os.path.isdir(fullpath):
            removeDirectory(
                path=fullpath,
                logger=tools_logger,
                ignore_errors=True,
                extra_recommendation=None,
            )


def makeNuitkaSourceDistribution(formats=None, sign=True):
    # spell-checker: ignore bztar,gztar
    if formats is None:
        formats = ("bztar", "gztar", "zip")

    checkAtHome()

    python = findInstalledPython(
        python_versions=("3.10", "3.11", "3.12"),
        module_specs=None,
    )

    cleanupSourceDistributionState()

    # Avoid strange permissions in archive
    os.system("umask 0022 && chmod -R a+rX .")

    filenames = []

    with withVirtualenv(
        "venv_nuitka",
        logger=tools_logger,
        style="blue",
        delete=False,
        python=python.getPythonExe(),
    ) as venv:
        # TODO: Should use a newer one, and predict more compliant filename, and
        # release 75.3.0 ought to work fine.
        venv.runCommand("python -m pip install -U 'setuptools==75.3.0'")
        venv.runCommand(
            "python setup.py sdist --formats=%s" % ",".join(formats), keep_cwd=True
        )

    # Cleanup the build directory, not needed.
    shutil.rmtree("build", ignore_errors=True)

    nuitka_version = getNuitkaVersionTuple()

    filename = _makeDistFilenameBase(nuitka_version) + ".tar.gz"
    assert os.path.exists(filename), filename

    # Delete requires.txt as it confuses poetry and potentially other tools
    assert os.system("gunzip -9 %s" % filename) == 0
    os.system(
        "tar --wildcards --delete --file %s nuitka-*/Nuitka.egg-info/requires.txt"
        % filename[:-3]
    )

    assert os.system("gzip -9 %s" % filename[:-3]) == 0
    assert os.path.exists(filename), filename

    assert os.system("chmod 644 %s" % filename) == 0

    filenames.append(filename)

    filename = _makeDistFilenameBase(nuitka_version) + ".tar.bz2"

    if os.path.exists(filename):
        # Delete requires.txt as it confuses poetry and potentially other tools,
        # spell-checker: ignore bunzip2
        assert os.system("bunzip2 %s" % filename) == 0
        os.system(
            "tar --wildcards --delete --file %s nuitka-*/Nuitka.egg-info/requires.txt"
            % filename[:-4]
        )
        assert os.system("bzip2 -9 %s" % filename[:-4]) == 0
        assert os.path.exists(filename), filename

        filenames.append(filename)

    filename = _makeDistFilenameBase(nuitka_version) + ".zip"

    if os.path.exists(filename):
        # Delete requires.txt as it confuses poetry and potentially other tools
        os.system("zip -d %s nuitka-*/Nuitka.egg-info/requires.txt" % filename)
        assert os.path.exists(filename), filename

        filenames.append(filename)

    for filename in filenames:
        assert os.system("chmod 644 %s" % filename) == 0

        if sign:
            assert (
                os.system("gpg --local-user 2912B99C --detach-sign %s" % filename) == 0
            )

    tools_logger.info("Created source distribution as '%s'." % filenames[0])
    return filenames


def syncTime():
    """Update WSL clock just in case, it can be skewed."""
    if isLinux() and "Microsoft" in getFileContents("/proc/version"):
        os.system("sudo hwclock -s")  # spell-checker: ignore hwclock


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        https://www.gnu.org/licenses/agpl-3.0.txt
#
#     See also: "Nuitka Runtime Library Exception, Version 1.0" in file
#     "LICENSE-RUNTIME.txt" for additional permissions granted under Section 7.
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
